
class RotorSE_Example1b():

	def execute(self):

		import numpy as np
		import os
		from openmdao.api import IndepVarComp, Component, Group, Problem, Brent, ScipyGMRES
		from rotorse.rotor_aeropower import RotorAeroPower
		from rotorse.rotor_geometry import RotorGeometry, NREL5MW, DTU10MW, NINPUT
		from rotorse import RPM2RS, RS2RPM, TURBULENCE_CLASS, DRIVETRAIN_TYPE, TURBINE_CLASS



		myref = DTU10MW()
		rotor = Problem()

		npts_coarse_power_curve = 20 # (Int): number of points to evaluate aero analysis at
		npts_spline_power_curve = 200  # (Int): number of points to use in fitting spline to power curve

		rotor.root = RotorAeroPower(myref, npts_coarse_power_curve, npts_spline_power_curve)
		rotor.setup()

		# === blade grid ===
		rotor['hubFraction'] = 0.023785  # (Float): hub location as fraction of radius
		rotor['bladeLength'] = 96.7  # (Float, m): blade length (if not precurved or swept) otherwise length of blade before curvature
		rotor['precone'] = 4.  # (Float, deg): precone angle
		rotor['tilt'] = 6.0  # (Float, deg): shaft tilt
		rotor['yaw'] = 0.0  # (Float, deg): yaw error
		rotor['nBlades'] = 3  # (Int): number of blades
		rotor['r_max_chord'] = 0.2366  # (Float): location of max chord on unit radius

		# Control points
		rotor['chord_in'] = np.array([4.6, 4.869795, 5.990629, 3.00785428, 0.0962])  # (Array, m): chord at control points. defined at hub, then at linearly spaced locations from r_max_chord to tip
		rotor['theta_in'] = np.array([14.5, 12.874, 6.724, -0.03388039, -0.037])  # (Array, deg): twist at control points.  defined at linearly spaced locations from r[idx_cylinder] to tip
		rotor['precurve_in'] = np.array([-0., -0.054497, -0.175303, -0.84976143, -6.206217])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
		rotor['presweep_in'] = np.array([0., 0., 0., 0., 0.])  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
		rotor['sparT_in'] = np.array([0.03200042, 0.07038508, 0.08515644, 0.07777004, 0.01181032])  # (Array, m): spar cap thickness parameters
		rotor['teT_in'] = np.array([0.04200055, 0.08807739, 0.05437378, 0.01610219, 0.00345225])  # (Array, m): trailing-edge thickness parameters
		# ------------------

		# === atmosphere ===
		rotor['analysis.rho'] = 1.225  # (Float, kg/m**3): density of air
		rotor['analysis.mu'] = 1.81206e-5  # (Float, kg/m/s): dynamic viscosity of air
		rotor['hub_height'] = 119. # (Float, m)
		rotor['analysis.shearExp'] = 0.25  # (Float): shear exponent
		rotor['turbine_class'] = TURBINE_CLASS['I']  # (Enum): IEC turbine class
		rotor['cdf_reference_height_wind_speed'] = 119.   # (Float): reference hub height for IEC wind speed (used in CDF calculation)
		# ----------------------

		# === control ===
		rotor['control_Vin'] = 4.0  # (Float, m/s): cut-in wind speed
		rotor['control_Vout'] = 25.0  # (Float, m/s): cut-out wind speed
		rotor['control_ratedPower'] = 10.e6  # (Float, W): rated power
		rotor['control_minOmega'] = 6.0  # (Float, rpm): minimum allowed rotor rotation speed
		rotor['control_maxOmega'] = 8.88766  # (Float, rpm): maximum allowed rotor rotation speed
		rotor['control_tsr'] = 10.58  # (Float): tip-speed ratio in Region 2 (should be optimized externally)
		rotor['control_pitch'] = 0.0  # (Float, deg): pitch angle in region 2 (and region 3 for fixed pitch machines)
		# ----------------------

		# === aero and structural analysis options ===
		rotor['nSector'] = 4  # (Int): number of sectors to divide rotor face into in computing thrust and power
		rotor['AEP_loss_factor'] = 1.0  # (Float): availability and other losses (soiling, array, etc.)
		rotor['drivetrainType'] = DRIVETRAIN_TYPE['GEARED']  # (Enum)
		# ----------------------

		# === run and outputs ===
		rotor.run()

		print 'AEP =', rotor['AEP']
		print 'diameter =', rotor['diameter']
		print 'ratedConditions.V =', rotor['rated_V']
		print 'ratedConditions.Omega =', rotor['rated_Omega']
		print 'ratedConditions.pitch =', rotor['rated_pitch']
		print 'ratedConditions.T =', rotor['rated_T']
		print 'ratedConditions.Q =', rotor['rated_Q']
		#for io in rotor.root.unknowns:
		#    print(io + ' ' + str(rotor.root.unknowns[io]))



		import matplotlib.pyplot as plt
		plt.figure()
		plt.plot(rotor['V'], rotor['P']/1e6)
		plt.xlabel('wind speed (m/s)')
		plt.xlabel('power (W)')
		plt.show()



		# outpath = '..\..\..\docs\images'

		# # Power Curve
		# f, ax = plt.subplots(1,1,figsize=(5.3, 4))
		# ax.plot(rotor['V'], rotor['P']/1e6)
		# ax.set(xlabel='Wind Speed (m/s)' , ylabel='Power (MW)')
		# ax.set_ylim([0, 10.3])
		# ax.set_xlim([0, 25])
		# f.tight_layout()
		# ax.spines['right'].set_visible(False)
		# ax.spines['top'].set_visible(False)
		# f.savefig(os.path.join(outpath,'power_curve_dtu10mw.png'))
		# f.savefig(os.path.join(outpath,'power_curve_dtu10mw.pdf'))

		# # Chord
		# fc, axc = plt.subplots(1,1,figsize=(5.3, 4))
		# rc_c = np.r_[0.0, myref.r_cylinder, np.linspace(rotor['r_max_chord'], 1.0, NINPUT-2)]
		# r = (rotor['spline.r_pts'] - rotor['spline.Rhub'])/rotor['bladeLength']
		# axc.plot(r, rotor['spline.chord'], c='k')
		# axc.plot(rc_c, rotor['chord_in'], '.', c='k')
		# for i, (x, y) in enumerate(zip(rc_c, rotor['chord_in'])):
		#     txt = '$c_%d$' % i
		#     if i<=1:
		#         axc.annotate(txt, (x,y), xytext=(x+0.01,y-0.4), textcoords='data')
		#     else:
		#         axc.annotate(txt, (x,y), xytext=(x+0.01,y+0.2), textcoords='data')
		# axc.set(xlabel='Blade Fraction, $r/R$' , ylabel='Chord (m)')
		# axc.set_ylim([0, 7])
		# axc.set_xlim([0, 1.1])
		# fc.tight_layout()
		# axc.spines['right'].set_visible(False)
		# axc.spines['top'].set_visible(False)
		# fc.savefig(os.path.join(outpath,'chord_dtu10mw.png'))
		# fc.savefig(os.path.join(outpath,'chord_dtu10mw.pdf'))

		# # Twist
		# ft, axt = plt.subplots(1,1,figsize=(5.3, 4))
		# rc_t = rc_c#np.linspace(myref.r_cylinder, 1.0, NINPUT)
		# r = (rotor['spline.r_pts'] - rotor['spline.Rhub'])/rotor['bladeLength']
		# axt.plot(r, rotor['spline.theta'], c='k')
		# axt.plot(rc_t, rotor['theta_in'], '.', c='k')
		# for i, (x, y) in enumerate(zip(rc_t, rotor['theta_in'])):
		#     txt = '$\Theta_%d$' % i
		#     axt.annotate(txt, (x,y), xytext=(x+0.01,y+0.1), textcoords='data')
		# axt.set(xlabel='Blade Fraction, $r/R$' , ylabel='Twist ($\deg$)')
		# axt.set_ylim([-1, 15])
		# axt.set_xlim([0, 1.1])
		# ft.tight_layout()
		# axt.spines['right'].set_visible(False)
		# axt.spines['top'].set_visible(False)
		# ft.savefig(os.path.join(outpath,'theta_dtu10mw.png'))
		# ft.savefig(os.path.join(outpath,'theta_dtu10mw.pdf'))


		# plt.show()



if __name__ == "__main__":

	rotor = RotorSE_Example1b()
	rotor.execute()


import unittest
import pytest
import sys

import wisdem.test.test_assemblies as test_assemblies
import wisdem.test.test_airfoilprep as test_airfoilprep
import wisdem.test.test_ccblade as test_ccblade
import wisdem.test.test_commonse as test_commonse
import wisdem.test.test_drivetrainse as test_drivetrainse
import wisdem.test.test_floatingse as test_floatingse
import wisdem.test.test_nrelcsm as test_nrelcsm
import wisdem.test.test_pbeam as test_pbeam
import wisdem.test.test_plant_financese as test_plant_financese
import wisdem.test.test_pyframe3dd as test_pyframe3dd
import wisdem.test.test_pymap as test_pymap
import wisdem.test.test_rotorse as test_rotorse
import wisdem.test.test_towerse as test_towerse
import wisdem.test.test_turbinecostsse as test_turbinecostsse
import wisdem.test.test_wisdem as test_wisdem
import wisdem.test.test_landbosse as test_landbosse

def suite():
    suite = unittest.TestSuite( (
        test_assemblies.test_all.suite(),
        test_airfoilprep.test_all.suite(),
        test_ccblade.test_all.suite(),
        test_commonse.test_all.suite(),
        #test_drivetrainse.test_all.suite(),
        test_floatingse.test_all.suite(),
        #test_nrelcsm.test_all.suite(),
        test_pbeam.test_all.suite(),
        test_plant_financese.test_all.suite(),
        test_pyframe3dd.test_all.suite(),
        #test_pymap.test_all.suite(),
        test_rotorse.test_all.suite(),
        test_towerse.test_all.suite(),
        test_turbinecostsse.test_all.suite()
        #test_wisdem.test_all.suite()                                 
    ) )
    return suite

valid_tests = ['test_orbit',
               'test_landbosse',
               'test_assemblies',
               'test_airfoilprep',
               'test_ccblade',
               'test_commonse',
               'test_floatingse',
               'test_pbeam',
               'test_plant_financese',
               'test_pyframe3dd',
               'test_towerse',
               #'test_drivetrainse',
               #'test_nrelcsm',
               #'test_pymap',
               'test_rotorse',
               #'test_wisdem',
               'test_turbinecostsse']

if __name__ == '__main__':
    #unittest.TextTestRunner().run(suite())
    ret = pytest.main(valid_tests)
    sys.exit(ret)



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.plant_financese.plant_finance as pf
from openmdao.api import Problem, Group

class TestPlantFinance(unittest.TestCase):
    def setUp(self):
        self.inputs = {}
        self.discrete_inputs = {}
        self.outputs = {}

        self.inputs['machine_rating'] = 1000.0
        self.inputs['tcc_per_kW'] = 1.2e3
        self.inputs['bos_per_kW'] = 7.7e3
        self.inputs['opex_per_kW'] = 7e2
        self.inputs['fixed_charge_rate'] = 0.12
        self.inputs['turbine_aep'] = 1.6e7
        self.inputs['park_aep'] = 1.6e7*50
        self.inputs['wake_loss_factor'] = 0.15

        self.discrete_inputs['turbine_number'] = 50
        
        self.mypfin = pf.PlantFinance()

    def testRun(self):
        # Park AEP way
        self.mypfin.compute(self.inputs, self.outputs, self.discrete_inputs, {})

        lcoe = 1e3*50*(0.12*(1.2e3+7.7e3) + 7e2) / (1.6e7*50.0)
        self.assertEqual(self.outputs['lcoe'], lcoe)

        # Wake loss way
        self.inputs['park_aep'] = 0.0
        self.mypfin.compute(self.inputs, self.outputs, self.discrete_inputs, {})

        lcoe = 1e3*50*(0.12*(1.2e3+7.7e3) + 7e2) / (1.6e7*50.0*(1-0.15))
        self.assertEqual(self.outputs['lcoe'], lcoe)


    def testDerivatives(self):
        prob = Problem()
        root = prob.model = Group()
        root.add_subsystem('pf', pf.PlantFinance(), promotes=['*'])
        prob.setup()
        for k in self.inputs.keys(): prob[k] = self.inputs[k]        
        for k in self.discrete_inputs.keys(): prob[k] = self.discrete_inputs[k]        
        #prob.check_partials()
        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestPlantFinance))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())
    
    



import numpy as np
import numpy.testing as npt
import unittest
import wisdem.commonse.vertical_cylinder as vc
from wisdem.commonse.utilities import nodal2sectional

npts = 100
myones = np.ones((npts,))

class TestDiscretization(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params['section_height'] = np.arange(1,5)
        self.params['diameter'] = 5.0 * np.ones(5)
        self.params['wall_thickness'] = 0.05 * np.ones(4)
        self.params['foundation_height'] = 0.0

    def testRefine2(self):
        mydis = vc.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns['z_param'], np.array([0.0, 1.0, 3.0, 6.0, 10.0]))
        npt.assert_array_equal(self.unknowns['z_full'], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]))
        npt.assert_array_equal(self.unknowns['d_full'], 5.0)
        npt.assert_array_equal(self.unknowns['t_full'], 0.05)

    def testFoundation(self):
        self.params['foundation_height'] = -30.0
        mydis = vc.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns['z_param'], np.array([0.0, 1.0, 3.0, 6.0, 10.0])-30.0)
        npt.assert_array_equal(self.unknowns['z_full'], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0])-30.0)
        npt.assert_array_equal(self.unknowns['d_full'], 5.0)
        npt.assert_array_equal(self.unknowns['t_full'], 0.05)

    def testRefine3(self):
        mydis = vc.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        for k in self.unknowns['z_param']:
            self.assertIn(k, self.unknowns['z_full'])
            
    def testRefineInterp(self):
        self.params['diameter'] = np.array([5.0, 5.0, 6.0, 7.0, 7.0])
        self.params['wall_thickness'] = 1e-2 * np.array([5.0, 5.0, 6.0, 7.0])
        mydis = vc.CylinderDiscretization(nPoints=5, nRefine=2)
        mydis.compute(self.params, self.unknowns)
        npt.assert_array_equal(self.unknowns['z_param'], np.array([0.0, 1.0, 3.0, 6.0, 10.0]))
        npt.assert_array_equal(self.unknowns['z_full'], np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.5, 6.0, 8.0, 10.0]))
        npt.assert_array_equal(self.unknowns['d_full'], np.array([5.0, 5.0, 5.0, 5.5, 6.0, 6.5, 7.0, 7.0, 7.0]))
        npt.assert_array_equal(self.unknowns['t_full'], 1e-2*np.array([5.0, 5.0, 5.0, 5.0, 6.0, 6.0, 7.0, 7.0]))

                               
class TestMass(unittest.TestCase):
    def setUp(self):
        self.params = {}
        self.unknowns = {}
        self.resid = None

        self.params['d_full'] = 2.0*10.0*myones
        self.params['t_full'] = 0.5*np.ones((npts-1,))
        self.params['z_full'] = np.linspace(0, 50.0, npts)
        self.params['material_density'] = 5.0
        self.params['outfitting_factor'] = 1.5
        self.params['material_cost_rate'] = 1.5
        self.params['labor_cost_rate'] = 1.0
        self.params['painting_cost_rate'] = 10.0
        
        self.cm = vc.CylinderMass(nPoints=npts)

    def testRegular(self):
        # Straight column
        self.cm.compute(self.params, self.unknowns)

        expect = np.pi*(10.**2 - 9.5**2)*5.0*1.5*(50.0/(npts-1))
        m = expect*(npts-1)
        Iax = 0.5*m*(10.**2 + 9.5**2)
        Ix = (1/12.)*m*(3*(10.**2 + 9.5**2) + 50*50) + m*25*25 # parallel axis on last term
        z_avg,_ = nodal2sectional(self.params['z_full'])
        self.assertAlmostEqual(self.unknowns['mass'].sum(), m)
        npt.assert_almost_equal(self.unknowns['mass'], expect)
        npt.assert_almost_equal(self.unknowns['section_center_of_mass'], z_avg)
        self.assertAlmostEqual(self.unknowns['center_of_mass'], 25.0)
        npt.assert_almost_equal(self.unknowns['I_base'], [Ix, Ix, Iax, 0.0, 0.0, 0.0], decimal=5)

        '''
    def testFrustum(self):
        # Frustum shell
        self.params['t_full'] = np.array([0.5, 0.4, 0.3])
        self.params['d_full'] = 2*np.array([10.0, 8.0, 6.0])
        self.wave.compute(self.params, self.unknowns)

        expect = np.pi/3.0*5.0*1.5*np.array([20.0, 30.0])*np.array([9.75*1.4+7.8*1.3, 7.8*1.1+5.85*1.0])
        m = expect*(npts-1)
        self.assertAlmostEqual(self.unknowns['mass'].sum(), m)
        npt.assert_almost_equal(self.unknowns['mass'].sum(), expect)
        '''

        
def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestDiscretization))
    suite.addTest(unittest.makeSuite(TestMass))
    return suite

if __name__ == '__main__':
    unittest.TextTestRunner().run(suite())



#!/usr/bin/env python
# encoding: utf-8
"""
csystem.py

Created by Andrew Ning on 2/21/2012.
Copyright (c) NREL. All rights reserved.
"""

from __future__ import print_function
import numpy as np


def rotMat_z(a):
    return np.asmatrix( np.array([[np.cos(a), -np.sin(a), 0.0],
                                  [np.sin(a),  np.cos(a), 0.0],
                                  [0.0,        0.0,       1.0]]) )

def rotMat_y(a):
    return np.asmatrix( np.array([[ np.cos(a), 0.0, np.sin(a)],
                                  [0.0,        1.0,       0.0],
                                  [-np.sin(a), 0.0, np.cos(a)] ]) )

def rotMat_x(a):
    return np.asmatrix( np.array([[1.0, 0.0,              0.0],
                                  [0.0, np.cos(a), -np.sin(a)],
                                  [0.0, np.sin(a),  np.cos(a)]]) )

# # !!!!!!!! NOTE ALL ANGLES SHOULD BE PASSED IN DEGREES !!!!!!!!!!!!!!
# def _rotateAboutZ(x, y, z, theta):
#     """
#     x X y = z.  rotate c.s. about z, +theta
#     all angles in degrees
#     """

#     theta = np.radians(theta)
#     c = np.cos(theta)
#     s = np.sin(theta)

#     xnew = x*c + y*s
#     ynew = -x*s + y*c

#     return xnew, ynew, z


# def _derivatives(obj, xstring, ystring, zstring, theta, rotateOtherWay):

#     x = getattr(obj, xstring)
#     y = getattr(obj, ystring)

#     xstring = 'd' + xstring
#     ystring = 'd' + ystring
#     zstring = 'd' + zstring

#     thetaM = 1.0
#     if rotateOtherWay:
#         thetaM = -1.0
#     theta *= thetaM

#     theta = np.radians(theta)
#     c = np.cos(theta)
#     s = np.sin(theta)

#     one = np.ones_like(x)
#     zero = np.zeros_like(x)

#     dxnew = {}
#     dxnew[xstring] = c
#     dxnew[ystring] = s
#     dxnew[zstring] = zero
#     dxnew['dtheta'] = (-x*s + y*c)*np.radians(thetaM)

#     dynew = {}
#     dynew[xstring] = -s
#     dynew[ystring] = c
#     dynew[zstring] = zero
#     dynew['dtheta'] = (-x*c - y*s)*np.radians(thetaM)

#     dznew = {}
#     dznew[xstring] = zero
#     dznew[ystring] = zero
#     dznew[zstring] = one
#     dznew['dtheta'] = zero

#     return dxnew, dynew, dznew




class DirectionVector(object):
    """Handles rotation of direction vectors to appropriate coordinate systems.
    All angles must be in degrees.

    """

    def __init__(self, x, y, z, dx=None, dy=None, dz=None):
        """3-Dimensional vector that depends on direction only (not position).

        Parameters
        ----------
        x : float or ndarray
            x-direction of vector(s)
        y : float or ndarray
            y-direction of vector(s)
        z : float or ndarray
            z-direction of vector(s)

        """

        self.x = np.array(x)
        self.y = np.array(y)
        self.z = np.array(z)

        if dx is None:
            dx = {}
            dx['dx'] = np.ones_like(self.x)
            dx['dy'] = np.zeros_like(self.y)
            dx['dz'] = np.zeros_like(self.z)

            dy = {}
            dy['dx'] = np.zeros_like(self.x)
            dy['dy'] = np.ones_like(self.y)
            dy['dz'] = np.zeros_like(self.z)

            dz = {}
            dz['dx'] = np.zeros_like(self.x)
            dz['dy'] = np.zeros_like(self.y)
            dz['dz'] = np.ones_like(self.z)

        self.dx = dx
        self.dy = dy
        self.dz = dz



    @classmethod
    def fromArray(cls, array):
        """initialize with NumPy array

        Parameters
        ----------
        array : ndarray
            construct DirectionVector using array of size 3

        """

        return cls(array[0], array[1], array[2])


    def toArray(self):
        """convert DirectionVector to NumPy array

        Returns
        -------
        array : ndarray
            NumPy array in order x, y, z containing DirectionVector data

        """

        return np.array([self.x, self.y, self.z])



    def _rotateAboutZ(self, xstring, ystring, zstring, theta, thetaname, reverse=False):
        """
        x X y = z.  rotate c.s. about z, +theta
        all angles in degrees
        """

        thetaM = 1.0
        if reverse:
            thetaM = -1.0

        x = getattr(self, xstring)
        y = getattr(self, ystring)
        z = getattr(self, zstring)
        dx = getattr(self, 'd' + xstring)
        dy = getattr(self, 'd' + ystring)
        dz = getattr(self, 'd' + zstring)

        theta = np.radians(theta * thetaM)
        c = np.cos(theta)
        s = np.sin(theta)

        xnew = x*c + y*s
        ynew = -x*s + y*c
        znew = z

        angles = []
        for key in dx.keys():
            if not key in ['dx', 'dy', 'dz']:
                angles.append(key)

        dxnew = {}
        dxnew['dx'] = dx['dx']*c + dy['dx']*s
        dxnew['dy'] = dx['dy']*c + dy['dy']*s
        dxnew['dz'] = dx['dz']*c + dy['dz']*s
        dxnew['d' + thetaname] = (-x*s + y*c)*np.radians(thetaM)
        for dangle in angles:
            dxnew[dangle] = dx[dangle]*c + dy[dangle]*s

        dynew = {}
        dynew['dx'] = -dx['dx']*s + dy['dx']*c
        dynew['dy'] = -dx['dy']*s + dy['dy']*c
        dynew['dz'] = -dx['dz']*s + dy['dz']*c
        dynew['d' + thetaname] = (-x*c - y*s)*np.radians(thetaM)
        for dangle in angles:
            dynew[dangle] = -dx[dangle]*s + dy[dangle]*c

        dznew = {}
        dznew['dx'] = dz['dx']*np.ones_like(theta)  # multiply by ones just to get right size in case of float
        dznew['dy'] = dz['dy']*np.ones_like(theta)
        dznew['dz'] = dz['dz']*np.ones_like(theta)
        dznew['d' + thetaname] = np.zeros_like(theta)
        for dangle in angles:
            dznew[dangle] = dz[dangle]


        return xnew, ynew, znew, dxnew, dynew, dznew





    def windToInertial(self, beta):
        """Rotates from wind-aligned to inertial

        Parameters
        ----------
        beta : float (deg)
            wind angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the inertial coordinate system

        """
        # xi, yi, zi = _rotateAboutZ(self.x, self.y, self.z, -beta)
        # return DirectionVector(xi, yi, zi)

        xw, yw, zw, dxw, dyw, dzw = self._rotateAboutZ('x', 'y', 'z', beta, 'beta', reverse=True)
        return DirectionVector(xw, yw, zw, dxw, dyw, dzw)


    def inertialToWind(self, beta):
        """Rotates from inertial to wind-aligned

        Parameters
        ----------
        beta : float (deg)
            wind angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the wind-aligned coordinate system

        """
        # xw, yw, zw = _rotateAboutZ(self.x, self.y, self.z, beta)
        # return DirectionVector(xw, yw, zw)

        xw, yw, zw, dxw, dyw, dzw = self._rotateAboutZ('x', 'y', 'z', beta, 'beta')
        return DirectionVector(xw, yw, zw, dxw, dyw, dzw)


    def yawToWind(self, Psi):
        """Rotates from yaw-aligned to wind-aligned

        Parameters
        ----------
        Psi : float (deg)
            yaw angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the wind-aligned coordinate system

        """
        # xw, yw, zw = _rotateAboutZ(self.x, self.y, self.z, -Psi)
        # return DirectionVector(xw, yw, zw)

        xw, yw, zw, dxw, dyw, dzw = self._rotateAboutZ('x', 'y', 'z', Psi, 'yaw', reverse=True)
        return DirectionVector(xw, yw, zw, dxw, dyw, dzw)


    def windToYaw(self, Psi):
        """Rotates from wind-aligned to yaw-aligned

        Parameters
        ----------
        Psi : float (deg)
            yaw angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the yaw-aligned coordinate system

        """
        # xy, yy, zy = _rotateAboutZ(self.x, self.y, self.z, Psi)
        # return DirectionVector(xy, yy, zy)

        xy, yy, zy, dxy, dyy, dzy = self._rotateAboutZ('x', 'y', 'z', Psi, 'yaw')
        return DirectionVector(xy, yy, zy, dxy, dyy, dzy)


    def hubToYaw(self, Theta, derivatives=False):
        """Rotates from hub-aligned to yaw-aligned

        Parameters
        ----------
        Theta : float (deg)
            tilt angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the yaw-aligned coordinate system

        """
        # zy, xy, yy = _rotateAboutZ(self.z, self.x, self.y, -Theta)
        # return DirectionVector(xy, yy, zy, deriv)

        zy, xy, yy, dzy, dxy, dyy = self._rotateAboutZ('z', 'x', 'y', Theta, 'tilt', reverse=True)
        return DirectionVector(xy, yy, zy, dxy, dyy, dzy)



    def yawToHub(self, Theta):
        """Rotates from yaw-aligned to hub-aligned

        Parameters
        ----------
        Theta : float (deg)
            tilt angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the hub-aligned coordinate system

        """
        # zh, xh, yh = _rotateAboutZ(self.z, self.x, self.y, Theta)
        # return DirectionVector(xh, yh, zh)

        zh, xh, yh, dzh, dxh, dyh = self._rotateAboutZ('z', 'x', 'y', Theta, 'tilt')
        return DirectionVector(xh, yh, zh, dxh, dyh, dzh)


    def hubToAzimuth(self, Lambda):
        """Rotates from hub-aligned to azimuth-aligned

        Parameters
        ----------
        Lambda : float or ndarray (deg)
            azimuth angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the azimuth-aligned coordinate system

        """
        # yz, zz, xz = _rotateAboutZ(self.y, self.z, self.x, Lambda)
        # return DirectionVector(xz, yz, zz)

        yz, zz, xz, dyz, dzz, dxz = self._rotateAboutZ('y', 'z', 'x', Lambda, 'azimuth')
        return DirectionVector(xz, yz, zz, dxz, dyz, dzz)


    def azimuthToHub(self, Lambda):
        """Rotates from azimuth-aligned to hub-aligned

        Parameters
        ----------
        Lambda : float or ndarray (deg)
            azimuth angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the hub-aligned coordinate system

        """
        # yh, zh, xh = _rotateAboutZ(self.y, self.z, self.x, -Lambda)
        # return DirectionVector(xh, yh, zh)

        yh, zh, xh, dyh, dzh, dxh = self._rotateAboutZ('y', 'z', 'x', Lambda, 'azimuth', reverse=True)
        return DirectionVector(xh, yh, zh, dxh, dyh, dzh)



    def azimuthToBlade(self, Phi):
        """Rotates from azimuth-aligned to blade-aligned

        Parameters
        ----------
        Phi : float (deg)
            precone angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the blade-aligned coordinate system

        """


        # zb, xb, yb = _rotateAboutZ(self.z, self.x, self.y, -Phi)
        # return DirectionVector(xb, yb, zb)

        zb, xb, yb, dzb, dxb, dyb = self._rotateAboutZ('z', 'x', 'y', Phi, 'precone', reverse=True)
        return DirectionVector(xb, yb, zb, dxb, dyb, dzb)



    def bladeToAzimuth(self, Phi):
        """Rotates from blade-aligned to azimuth-aligned

        Parameters
        ----------
        Phi : float (deg)
            precone angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the azimuth-aligned coordinate system


        """

        # za, xa, ya = _rotateAboutZ(self.z, self.x, self.y, Phi)
        # return DirectionVector(xa, ya, za)

        za, xa, ya, dza, dxa, dya = self._rotateAboutZ('z', 'x', 'y', Phi, 'precone')
        return DirectionVector(xa, ya, za, dxa, dya, dza)



    def airfoilToBlade(self, theta):
        """Rotates from airfoil-aligned to blade-aligned

        Parameters
        ----------
        theta : float (deg)
            twist angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the blade-aligned coordinate system

        """
        # xb, yb, zb = _rotateAboutZ(self.x, self.y, self.z, theta)
        # return DirectionVector(xb, yb, zb)

        xb, yb, zb, dxb, dyb, dzb = self._rotateAboutZ('x', 'y', 'z', theta, 'theta')
        return DirectionVector(xb, yb, zb, dxb, dyb, dzb)


    def bladeToAirfoil(self, theta):
        """Rotates from blade-aligned to airfoil-aligned

        Parameters
        ----------
        theta : float (deg)
            twist angle

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the airfoil-aligned coordinate system

        """

        # xa, ya, za = _rotateAboutZ(self.x, self.y, self.z, -theta)
        # return DirectionVector(xa, ya, za)

        xa, ya, za, dxa, dya, dza = self._rotateAboutZ('x', 'y', 'z', theta, 'theta', reverse=True)
        return DirectionVector(xa, ya, za, dxa, dya, dza)


    def airfoilToProfile(self):
        """Rotates from airfoil-aligned to profile

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the profile coordinate system

        """

        # return DirectionVector(self.y, self.x, self.z)

        return DirectionVector(self.y, self.x, self.z, self.dy, self.dx, self.dz)


    def profileToAirfoil(self):
        """Rotates from profile to airfoil-aligned

        Returns
        -------
        vector : DirectionVector
            a DirectionVector in the airfoil-aligned coordinate system

        """

        # return DirectionVector(self.y, self.x, self.z)

        return DirectionVector(self.y, self.x, self.z, self.dy, self.dx, self.dz)




    def cross(self, other):
        """cross product between two DirectionVectors

        Parameters
        ----------
        other : DirectionVector
            other vector to cross with

        Returns
        -------
        vector : DirectionVector
            vector = self X other

        """

        v1 = np.array([self.x, self.y, self.z]).transpose()
        v2 = np.array([other.x, other.y, other.z]).transpose()
        v = np.cross(v1, v2)

        if len(v.shape) > 1:
            return DirectionVector(v[:, 0], v[:, 1], v[:, 2])
        else:
            return DirectionVector(v[0], v[1], v[2])


    def cross_deriv(self, other, namea='a', nameb='b'):
        """defined only for floats for now"""

        # c = a X b
        a = self
        b = other

        dx = {}
        dx[namea] = np.array([0.0, b.z, -b.y])
        dx[nameb] = np.array([0.0, -a.z, a.y])

        dy = {}
        dy[namea] = np.array([-b.z, 0.0, b.x])
        dy[nameb] = np.array([a.z, 0.0, -a.x])

        dz = {}
        dz[namea] = np.array([b.y, -b.x, 0.0])
        dz[nameb] = np.array([-a.y, a.x, 0.0])

        return dx, dy, dz



    def cross_deriv_array(self, other, namea='a', nameb='b'):

        # c = a X b
        a = self
        b = other

        dx = {}
        dx['d' + namea + 'x'] = np.zeros_like(b.x)
        dx['d' + namea + 'y'] = b.z
        dx['d' + namea + 'z'] = -b.y
        dx['d' + nameb + 'x'] = np.zeros_like(a.x)
        dx['d' + nameb + 'y'] = -a.z
        dx['d' + nameb + 'z'] = a.y

        dy = {}
        dy['d' + namea + 'x'] = -b.z
        dy['d' + namea + 'y'] = np.zeros_like(b.y)
        dy['d' + namea + 'z'] = b.x
        dy['d' + nameb + 'x'] = a.z
        dy['d' + nameb + 'y'] = np.zeros_like(a.y)
        dy['d' + nameb + 'z'] = -a.x

        dz = {}
        dz['d' + namea + 'x'] = b.y
        dz['d' + namea + 'y'] = -b.x
        dz['d' + namea + 'z'] = np.zeros_like(b.z)
        dz['d' + nameb + 'x'] = -a.y
        dz['d' + nameb + 'y'] = a.x
        dz['d' + nameb + 'z'] = np.zeros_like(a.z)

        return dx, dy, dz




    def __neg__(self):
        """negate direction vector"""

        return DirectionVector(-self.x, -self.y, -self.z)


    def __add__(self, other):
        """add two DirectionVector objects (v1 = v2 + v3)"""

        if isinstance(other, DirectionVector):
            return DirectionVector(self.x + other.x, self.y + other.y, self.z + other.z)
        else:
            return DirectionVector(self.x + other, self.y + other, self.z + other)


    def __sub__(self, other):
        """subtract DirectionVector objects (v1 = v2 - v3)"""

        if isinstance(other, DirectionVector):
            return DirectionVector(self.x - other.x, self.y - other.y, self.z - other.z)
        else:
            return DirectionVector(self.x - other, self.y - other, self.z - other)


    def __iadd__(self, other):
        """add DirectionVector object to self (v1 += v2)"""

        if isinstance(other, DirectionVector):
            self.x += other.x
            self.y += other.y
            self.z += other.z
        else:
            self.x += other
            self.y += other
            self.z += other

        return self


    def __isub__(self, other):
        """subract DirectionVector object from self (v1 -= v2)"""

        if isinstance(other, DirectionVector):
            self.x -= other.x
            self.y -= other.y
            self.z -= other.z
        else:
            self.x -= other
            self.y -= other
            self.z -= other

        return self


    def __mul__(self, other):
        """multiply vector times a scalar or element by element muiltiply times another vector (v1 = alpha * v2 or v1 = v2 * v3)"""

        if isinstance(other, DirectionVector):
            return DirectionVector(self.x * other.x, self.y * other.y, self.z * other.z)
        else:
            return DirectionVector(self.x * other, self.y * other, self.z * other)


    def __div__(self, other):
        """divide vector by a scalar or element by element division with another vector (v1 = v2 / alpha or v1 = v2 / v3)"""

        if isinstance(other, DirectionVector):
            return DirectionVector(self.x / other.x, self.y / other.y, self.z / other.z)
        else:
            return DirectionVector(self.x / float(other), self.y / float(other), self.z / float(other))


    def __imul__(self, other):
        """multiply self times a scalar or element by element muiltiply times another vector (v1 *= alpha or v1 *= v2)"""

        if isinstance(other, DirectionVector):
            self.x *= other.x
            self.y *= other.y
            self.z *= other.z
        else:
            self.x *= other
            self.y *= other
            self.z *= other
        return self


    def __idiv__(self, other):
        """divide self by a scalar or element by element division with another vector (v1 /= alpha or v1 /= v2)"""
        if isinstance(other, DirectionVector):
            self.x /= other.x
            self.y /= other.y
            self.z /= other.z
        else:
            self.x /= float(other)
            self.y /= float(other)
            self.z /= float(other)
        return self


    def __str__(self):
        """print string representation"""

        return '{0}, {1}, {2}'.format(self.x, self.y, self.z)





# class Force(DirectionVector):

#     def __init__(self, Fx, Fy, Fz):
#         """Force vector (or any vector that depends only on orientation).

#         Parameters
#         ----------
#         Fx : float or ndarray
#             force in x-direction
#         Fy : float or ndarray
#             force in y-direction
#         Fz : float or ndarray
#             force in z-direction

#         """
#         super(DirectionVector, self).__init__(Fx, Fy, Fz)




# def _position(r):
#     return Position(r.x, r.y, r.z)


# class Position(DirectionVector):

#     def __init__(self, x, y, z):
#         """Position vector.

#         Parameters
#         ----------
#         x : float or ndarray
#             x-location of position(s)
#         y : float or ndarray
#             y-location of position(s)
#         z : float or ndarray
#             z-location of position(s)

#         """
#         super(Position, self).__init__(x, y, z)


#     def __translate(self, x, y, z):
#         return self - DirectionVector(x, y, z)


#     def yawToWind(self, Psi, height):
#         r = self.__translate(0.0, 0.0, -height)
#         rnew = r.yawToWind(Psi)

#         return _position(rnew)


#     def windToYaw(self, Psi, height):
#         r = self.__translate(0.0, 0.0, height)
#         rnew = r.windToYaw(Psi)

#         return _position(rnew)


#     def hubToYaw(self, Theta, length):
#         r = self.__translate(length, 0.0, 0.0)
#         rnew = r.hubToYaw(Theta)

#         return _position(rnew)


#     def yawToHub(self, Theta, length):
#         r = DirectionVector(self.x, self.y, self.z).yawToHub(Theta)
#         rnew = r.__translate(-length, 0.0, 0.0)

#         return _position(rnew)


#     def hubToAzimuth(self, Lambda, dy, dz):
#         r = self.__translate(0.0, dy, dz)
#         rnew = r.hubToAzimuth(Lambda)

#         return _position(rnew)

#     def azimuthToHub(self, Lambda, dy, dz):
#         r = DirectionVector(self.x, self.y, self.z).azimuthToHub(Lambda)
#         rnew = _position(r).__translate(0.0, -dy, -dz)

#         return _position(rnew)


#     def bladeToAzimuth(self, Phi):
#         r = DirectionVector(self.x, self.y, self.z).bladeToAzimuth(Phi)
#         return _position(r)


#     def profileToAirfoil(self, d):
#         r = self.__translate(d, 0.0, 0.0)
#         rnew = r.profileToAirfoil()

#         return _position(rnew)


#     def airfoilToProfile(self, d):
#         r = self.airfoilToProfile()
#         rnew = r.__translate(-d, 0.0, 0.0)

#         return _position(rnew)



# class Velocity(DirectionVector):

#     def __init__(self, Vx, Vy, Vz):
#         """Velocity vector.

#         Parameters
#         ----------
#         Vx : float or ndarray
#             velocity in x-direction
#         Vy : float or ndarray
#             velocity in y-direction
#         Vz : float or ndarray
#             velocity in z-direction

#         """
#         super(DirectionVector, self).__init__(Vx, Vy, Vz)


#     def hubToAzimuth(self, Lambda, Omega, yh, zh):

#         yh = np.array(yh)
#         zh = np.array(zh)
#         r = np.sqrt(yh**2 + zh**2)
#         beta = np.arctan2(zh, -yh)
#         Vy = Omega*r*np.sin(beta)
#         Vz = Omega*r*np.cos(beta)

#         V = self + Velocity(0, Vy, Vz)
#         Vnew = DirectionVector(V.x, V.y, V.z).hubToAzimuth(Lambda)

#         return Velocity(Vnew.x, Vnew.y, Vnew.z)


#     def azimuthToHub(self, Lambda, Omega, yh, zh):

#         yh = np.array(yh)
#         zh = np.array(zh)
#         r = np.sqrt(yh**2 + zh**2)
#         beta = np.arctan2(zh, -yh)
#         Vy = Omega*r*np.sin(beta)
#         Vz = Omega*r*np.cos(beta)

#         V = DirectionVector(self.x, self.y, self.z).azimuthToHub(Lambda)
#         Vnew = V - Velocity(0, Vy, Vz)

#         return Velocity(Vnew.x, Vnew.y, Vnew.z)





# def _FM(f, m):
#     return ForceMoment(f.x, f.y, f.z, m.x, m.y, m.z)


# class ForceMoment:

#     def __init__(self, Fx, Fy, Fz, Mx, My, Mz):
#         """Force vector and a corresponding moment vector.

#         Parameters
#         ----------
#         Fx : float or ndarray
#             x-direction of force(s)
#         Fy : float or ndarray
#             y-direction of force(s)
#         Fz : float or ndarray
#             z-direction of force(s)
#         Mx : float or ndarray
#             x-direction of moment(s)
#         My : float or ndarray
#             y-direction of moment(s)
#         Mz : float or ndarray
#             z-direction of moment(s)

#         """

#         self.F = DirectionVector(Fx, Fy, Fz)
#         self.M = DirectionVector(Mx, My, Mz)


#     def __translate(self, x, y, z):
#         r = DirectionVector(x, y, z)
#         M = self.M - r.cross(self.F)

#         return self.F, M


#     def windToInertial(self, beta):

#         Fnew = self.F.windToInertial(beta)
#         Mnew = self.M.windToInertial(beta)

#         return _FM(Fnew, Mnew)


#     def inertialToWind(self, beta):
#         """docstring"""

#         Fnew = self.F.inertialToWind(beta)
#         Mnew = self.M.inertialToWind(beta)

#         return _FM(Fnew, Mnew)


#     def yawToWind(self, psi, height):
#         """docstring"""

#         F, M = self.__translate(0.0, 0.0, -height)
#         Fnew = F.yawToWind(psi)
#         Mnew = M.yawToWind(psi)

#         return _FM(Fnew, Mnew)


#     def windToYaw(self, psi, height):
#         """docstring"""

#         F, M = self.__translate(0.0, 0.0, height)
#         Fnew = F.windToYaw(psi)
#         Mnew = M.windToYaw(psi)

#         return _FM(Fnew, Mnew)


#     def hubToYaw(self, Theta, length):

#         F, M = self.__translate(length, 0.0, 0.0)
#         print(self.F, F)
#         print(self.M, M)
#         Fnew = F.hubToYaw(Theta)
#         Mnew = M.hubToYaw(Theta)

#         return _FM(Fnew, Mnew)


#     def yawToHub(self, Theta, length):

#         F = self.F.yawToHub(Theta)
#         M = self.M.yawToHub(Theta)
#         Fnew, Mnew = _FM(F, M).__translate(-length, 0.0, 0.0)

#         return _FM(Fnew, Mnew)


#     def hubToAzimuth(self, Lambda, yh, zh):

#         F, M = self.__translate(0.0, yh, zh)
#         Fnew = F.hubToAzimuth(Lambda)
#         Mnew = M.hubToAzimuth(Lambda)

#         return _FM(Fnew, Mnew)


#     def azimuthToHub(self, Lambda, yh, zh):

#         F = self.F.azimuthToHub(Lambda)
#         M = self.M.azimuthToHub(Lambda)
#         Fnew, Mnew = _FM(F, M).__translate(0.0, -yh, -zh)

#         return _FM(Fnew, Mnew)


#     def azimuthToBlade(self, Phi):

#         Fnew = self.F.azimuthToBlade(Phi)
#         Mnew = self.M.azimuthToBlade(Phi)

#         return _FM(Fnew, Mnew)


#     def bladeToAzimuth(self, Phi):

#         Fnew = self.F.bladeToAzimuth(Phi)
#         Mnew = self.M.bladeToAzimuth(Phi)

#         return _FM(Fnew, Mnew)


#     def airfoilToBlade(self, theta):

#         Fnew = self.F.airfoilToBlade(theta)
#         Mnew = self.M.airfoilToBlade(theta)

#         return _FM(Fnew, Mnew)


#     def bladeToAirfoil(self, theta):

#         Fnew = self.F.bladeToAirfoil(theta)
#         Mnew = self.M.bladeToAirfoil(theta)

#         return _FM(Fnew, Mnew)


#     def profileToAirfoil(self, d):

#         F, M = self.__translate(d, 0.0, 0.0)
#         Fnew = F.profileToAirfoil()
#         Mnew = M.profileToAirfoil()

#         return _FM(Fnew, Mnew)


#     def airfoilToProfile(self, d):

#         F = self.F.airfoilToProfile()
#         M = self.M.airfoilToProfile()
#         Fnew, Mnew = _FM(F, M).__translate(-d, 0.0, 0.0)

#         return _FM(Fnew, Mnew)


#     def __iadd__(self, other):
#         self.F += other.F
#         self.M += other.M

#         return self

#     def __div__(self, other):
#         F = DirectionVector(self.F.x, self.F.y, self.F.z).__div__(other)
#         M = DirectionVector(self.M.x, self.M.y, self.M.z).__div__(other)

#         return _FM(F, M)


# def example():
#     # from twister.common import DirectionVector

#     # define a thrust load in the rotor-aligned coordinate system
#     T = 1.0e6
#     thrust = DirectionVector(T, 0.0, 0.0)  # thrust load in the x_r direction

#     # define angles (in degrees)
#     tilt = 10.0
#     yaw = 15.0

#     # translate to wind-aligned coordinate system (through an intermediate frame)
#     Fwind = thrust.rotorToYaw(tilt).yawToWind(yaw)

#     # print the components
#     print('Fwind components (x 10^6 N) =', Fwind.x/1e6, Fwind.y/1e6, Fwind.z/1e6)


if __name__ == '__main__':

    x = np.array([1.0, 2.0])
    y = np.array([1.3, 4.3])
    z = np.array([2.3, 2.3])
    a = DirectionVector(x, y, z)

    x = np.array([3.2, 1.5])
    y = np.array([2.1, 3.2])
    z = np.array([5.6, 7.7])
    b = DirectionVector(x, y, z)

    dx, dy, dz = a.cross_deriv_array(b)

    print(dx['dax'])
    print(dx['day'])
    print(dx['daz'])
    print(dx['dby'])

    exit()


    T1 = 5.0
    T2 = 4.0
    T3 = -3.0
    tilt = 20.0

    F = DirectionVector(T1, T2, T3).hubToYaw(tilt)

    Fp1 = DirectionVector(T1+1e-6, T2, T3).hubToYaw(tilt)
    Fp2 = DirectionVector(T1, T2+1e-6, T3).hubToYaw(tilt)
    Fp3 = DirectionVector(T1, T2, T3+1e-6).hubToYaw(tilt)
    Fp4 = DirectionVector(T1, T2, T3).hubToYaw(tilt+1e-6)

    dFx = F.dx
    dFy = F.dy
    dFz = F.dz
    # dFx, dFy, dFz = DirectionVector(T1, T2, T3).hubToYaw(tilt, derivatives=True)

    print(dFx['dx'])
    print((Fp1.x - F.x)/1e-6)
    print(dFx['dy'])
    print((Fp2.x - F.x)/1e-6)
    print(dFx['dz'])
    print((Fp3.x - F.x)/1e-6)
    print(dFx['dtilt'], 'theta')
    print((Fp4.x - F.x)/1e-6)


    print(dFy['dx'])
    print((Fp1.y - F.y)/1e-6)
    print(dFy['dy'])
    print((Fp2.y - F.y)/1e-6)
    print(dFy['dz'])
    print((Fp3.y - F.y)/1e-6)
    print(dFy['dtilt'], 'theta')
    print((Fp4.y - F.y)/1e-6)


    print(dFz['dx'])
    print((Fp1.z - F.z)/1e-6)
    print(dFz['dy'])
    print((Fp2.z - F.z)/1e-6)
    print(dFz['dz'])
    print((Fp3.z - F.z)/1e-6)
    print(dFz['dtilt'], 'theta')
    print((Fp4.z - F.z)/1e-6)


    print()
    print()
    print()
    yaw = 3.0
    F = DirectionVector(T1, T2, T3).hubToYaw(tilt).yawToWind(yaw)

    Fp1 = DirectionVector(T1+1e-6, T2, T3).hubToYaw(tilt).yawToWind(yaw)
    Fp2 = DirectionVector(T1, T2+1e-6, T3).hubToYaw(tilt).yawToWind(yaw)
    Fp3 = DirectionVector(T1, T2, T3+1e-6).hubToYaw(tilt).yawToWind(yaw)
    Fp4 = DirectionVector(T1, T2, T3).hubToYaw(tilt+1e-6).yawToWind(yaw)
    Fp5 = DirectionVector(T1, T2, T3).hubToYaw(tilt).yawToWind(yaw+1e-6)

    dFx = F.dx
    dFy = F.dy
    dFz = F.dz

    print(dFx['dx'])
    print((Fp1.x - F.x)/1e-6)
    print(dFx['dy'])
    print((Fp2.x - F.x)/1e-6)
    print(dFx['dz'])
    print((Fp3.x - F.x)/1e-6)
    print(dFx['dtilt'], 'theta')
    print((Fp4.x - F.x)/1e-6)
    print(dFx['dyaw'])
    print((Fp5.x - F.x)/1e-6)

    print(dFy['dx'])
    print((Fp1.y - F.y)/1e-6)
    print(dFy['dy'])
    print((Fp2.y - F.y)/1e-6)
    print(dFy['dz'])
    print((Fp3.y - F.y)/1e-6)
    print(dFy['dtilt'], 'theta')
    print((Fp4.y - F.y)/1e-6)
    print(dFy['dyaw'])
    print((Fp5.y - F.y)/1e-6)


    print(dFz['dx'])
    print((Fp1.z - F.z)/1e-6)
    print(dFz['dy'])
    print((Fp2.z - F.z)/1e-6)
    print(dFz['dz'])
    print((Fp3.z - F.z)/1e-6)
    print(dFz['dtilt'], 'theta')
    print((Fp4.z - F.z)/1e-6)
    print(dFz['dyaw'])
    print((Fp5.z - F.z)/1e-6)



    # x = DirectionVector(4.0, 5.0, 6.0)

    # print(x)

    # x = DirectionVector([4.0]*2, [5.0]*2, [6.0]*2)

    # print(x)

    # # p = Position(1.0, 2.0, 3.0)

    # # x = DirectionVector(4.0, 5.0, 6.0)

    # # print(p - x)

    # # y = p - x

    # # print(p.__class__)
    # # print(x.__class__)
    # # print(y.__class__)

    # # print(p.yawToWind(0.0, 10.0).yawToWind(0.0, 10.0))

    # # # example()





import numpy as np

from wisdem.floating.floating_turbine_instance import FloatingTurbineInstance, NPTS, vecOption
import offshorebos.wind_obos as wind_obos
from commonse.utilities import sectional2nodal

class TurbineSemiInstance(FloatingTurbineInstance):
    def __init__(self, refStr):
        super(TurbineSemiInstance, self).__init__(refStr)

        self.params['substructure'] =                 wind_obos.Substructure.SEMISUBMERSIBLE
        self.params['anchor'] =                       wind_obos.Anchor.DRAGEMBEDMENT
        self.params['turbInstallMethod'] =            wind_obos.TurbineInstall.INDIVIDUAL
        self.params['towerInstallMethod'] =           wind_obos.TowerInstall.ONEPIECE
        self.params['installStrategy'] =              wind_obos.InstallStrategy.PRIMARYVESSEL

        # Change scalars to vectors where needed
        self.check_vectors()


        
    def visualize(self, fname=None):
        fig = self.init_figure()

        self.draw_ocean(fig)

        mooringMat = self.prob['mooring_plot_matrix']
        self.draw_mooring(fig, mooringMat)

        pontoonMat = self.prob['plot_matrix']
        zcut = 1.0 + np.maximum( self.params['main_freeboard'], self.params['offset_freeboard'] )
        self.draw_pontoons(fig, pontoonMat, 0.5*self.params['pontoon_outer_diameter'], zcut)

        self.draw_column(fig, [0.0, 0.0], self.params['main_freeboard'], self.params['main_section_height'],
                           0.5*self.params['main_outer_diameter'], self.params['main_stiffener_spacing'])

        t_full = sectional2nodal(self.params['main_wall_thickness'])
        self.draw_ballast(fig, [0.0, 0.0], self.params['main_freeboard'], self.params['main_section_height'],
                          0.5*self.params['main_outer_diameter']-t_full,
                          self.params['main_permanent_ballast_height'], self.prob['variable_ballast_height'])

        if self.prob['main.buoyancy_tank_mass'] > 0.0:
            self.draw_buoyancy_tank(fig, [0.0, 0.0], self.params['main_freeboard'],
                                        self.params['main_section_height'],
                                        self.params['main_buoyancy_tank_location'],
                                        0.5*self.params['main_buoyancy_tank_diameter'],
                                        self.params['main_buoyancy_tank_height'])
        
        R_semi  = self.params['radius_to_offset_column']
        ncolumn = int(self.params['number_of_offset_columns'])
        angles = np.linspace(0, 2*np.pi, ncolumn+1)
        x = R_semi * np.cos( angles )
        y = R_semi * np.sin( angles )
        for k in range(ncolumn):
            self.draw_column(fig, [x[k], y[k]], self.params['offset_freeboard'], self.params['offset_section_height'],
                               0.5*self.params['offset_outer_diameter'], self.params['offset_stiffener_spacing'])

            t_full = sectional2nodal(self.params['offset_wall_thickness'])
            self.draw_ballast(fig, [x[k], y[k]], self.params['offset_freeboard'], self.params['offset_section_height'],
                              0.5*self.params['offset_outer_diameter']-t_full,
                              self.params['offset_permanent_ballast_height'], 0.0)

            if self.prob['off.buoyancy_tank_mass'] > 0.0:
                self.draw_buoyancy_tank(fig, [x[k], y[k]], self.params['offset_freeboard'],
                                            self.params['offset_section_height'],
                                            self.params['offset_buoyancy_tank_location'],
                                            0.5*self.params['offset_buoyancy_tank_diameter'],
                                            self.params['offset_buoyancy_tank_height'])
            
        self.draw_column(fig, [0.0, 0.0], self.params['hub_height'], self.params['tower_section_height'],
                         0.5*self.params['tower_outer_diameter'], None, (0.9,)*3)

        self.draw_rna(fig)
        
        self.set_figure(fig, fname)





from openmdao.api import Group, IndepVarComp, Problem, ExplicitComponent
from wisdem.floatingse.column import Column, ColumnGeometry
from wisdem.floatingse.substructure import Substructure, SubstructureGeometry
from wisdem.floatingse.loading import Loading
from wisdem.floatingse.map_mooring import MapMooring
from wisdem.towerse.tower import TowerLeanSE
import numpy as np

    
class FloatingSE(Group):

    def initialize(self):
        self.options.declare('nSection', default=4)
        self.options.declare('nTower', default=3)
        self.options.declare('nRefine', default=3)
        self.options.declare('topLevelFlag', default=True)

    def setup(self):

        nSection = self.options['nSection']
        nTower   = self.options['nTower']
        nRefine  = self.options['nRefine']
        topLevelFlag = self.options['topLevelFlag']
        nFullSec = nRefine*nSection+1
        nFullTow = nRefine*nTower  +1

        self.add_subsystem('tow', TowerLeanSE(nPoints=nTower+1, nFull=nFullTow, topLevelFlag=False), promotes=['material_density','tower_section_height',
                                                                        'tower_outer_diameter','tower_wall_thickness','tower_outfitting_factor',
                                                                        'tower_buckling_length','max_taper','min_d_to_t','rna_mass','rna_cg','rna_I',
                                                                        'tower_mass','tower_I_base','hub_height',#'tip_position','hub_cm','downwind',
                                                                        'material_cost_rate','labor_cost_rate','painting_cost_rate'])
        
        # Next do main and ballast columns
        # Ballast columns are replicated from same design in the components
        self.add_subsystem('main', Column(nSection=nSection, nFull=nFullSec, topLevelFlag=False), promotes=['water_density','E','nu','yield_stress','z0',
                                                                                                            'Uref','zref','shearExp','yaw','Uc','Hs','hmax','T','cd_usr','cm','loading',
                                                                 'max_draft','max_taper','min_d_to_t','gamma_f','gamma_b','foundation_height',
                                                                 'permanent_ballast_density','bulkhead_mass_factor','buoyancy_tank_mass_factor',
                                                                 'ring_mass_factor','column_mass_factor','outfitting_mass_fraction','ballast_cost_rate',
                                                                 'material_cost_rate','labor_cost_rate','painting_cost_rate','outfitting_cost_rate'])
        self.add_subsystem('off', Column(nSection=nSection, nFull=nFullSec, topLevelFlag=False), promotes=['water_density','E','nu','yield_stress','z0',
                                                                                                           'Uref','zref','shearExp','yaw','Uc','Hs','hmax','T','cd_usr','cm','loading',
                                                                'max_draft','max_taper','min_d_to_t','gamma_f','gamma_b','foundation_height',
                                                                'permanent_ballast_density','bulkhead_mass_factor','buoyancy_tank_mass_factor',
                                                                'ring_mass_factor','column_mass_factor','outfitting_mass_fraction','ballast_cost_rate',
                                                                'material_cost_rate','labor_cost_rate','painting_cost_rate','outfitting_cost_rate'])

        # Run Semi Geometry for interfaces
        self.add_subsystem('sg', SubstructureGeometry(nFull=nFullSec, nFullTow=nFullTow), promotes=['*'])

        # Next run MapMooring
        self.add_subsystem('mm', MapMooring(), promotes=['*'])
        
        # Add in the connecting truss
        self.add_subsystem('load', Loading(nFull=nFullSec, nFullTow=nFullTow), promotes=['*'])

        # Run main Semi analysis
        self.add_subsystem('subs', Substructure(nFull=nFullSec, nFullTow=nFullTow), promotes=['*'])

        # Independent variables that may be duplicated at higher levels of aggregation
        if topLevelFlag:
            sharedIndeps = IndepVarComp()
            sharedIndeps.add_output('hub_height', 0.0, units='m')
            sharedIndeps.add_output('material_density', 0.0, units='kg/m**3')
            sharedIndeps.add_output('air_density', 0.0, units='kg/m**3')
            sharedIndeps.add_output('air_viscosity', 0.0, units='kg/m/s')
            sharedIndeps.add_output('shearExp', 0.0)
            sharedIndeps.add_output('wind_reference_height', 0.0, units='m')
            sharedIndeps.add_output('wind_reference_speed', 0.0, units='m/s')
            sharedIndeps.add_output('wind_z0', 0.0, units='m')
            sharedIndeps.add_output('wind_beta', 0.0, units='deg')
            sharedIndeps.add_output('cd_usr', -1.0)
            sharedIndeps.add_output('cm', 0.0)
            sharedIndeps.add_output('Uc', 0.0, units='m/s')
            sharedIndeps.add_output('water_depth', 0.0, units='m')
            sharedIndeps.add_output('water_density', 0.0, units='kg/m**3')
            sharedIndeps.add_output('water_viscosity', 0.0, units='kg/m/s')
            sharedIndeps.add_output('significant_wave_height', 0.0, units='m')
            sharedIndeps.add_output('significant_wave_period', 0.0, units='s')
            sharedIndeps.add_output('wave_beta', 0.0, units='deg')
            sharedIndeps.add_output('wave_z0', 0.0, units='m')
            sharedIndeps.add_output('yaw', 0.0, units='deg')
            sharedIndeps.add_output('E', 0.0, units='N/m**2')
            sharedIndeps.add_output('G', 0.0, units='N/m**2')
            sharedIndeps.add_output('nu', 0.0)
            sharedIndeps.add_output('yield_stress', 0.0, units='N/m**2')
            sharedIndeps.add_output('gamma_f', 0.0)
            sharedIndeps.add_output('gamma_m', 0.0)
            sharedIndeps.add_output('gamma_n', 0.0)
            sharedIndeps.add_output('gamma_b', 0.0)
            sharedIndeps.add_output('gamma_fatigue', 0.0)
            sharedIndeps.add_output('DC', 0.0)
            sharedIndeps.add_discrete_output('shear', True)
            sharedIndeps.add_discrete_output('geom', False)
            sharedIndeps.add_discrete_output('nM', 2)
            sharedIndeps.add_discrete_output('Mmethod', 1)
            sharedIndeps.add_discrete_output('lump', 0)
            sharedIndeps.add_output('tol', 0.0)
            sharedIndeps.add_output('shift', 0.0)
            sharedIndeps.add_output('life', 0.0)
            sharedIndeps.add_output('m_SN', 0.0)
            sharedIndeps.add_output('rna_mass', 0.0, units='kg')
            sharedIndeps.add_output('rna_I', np.zeros(6), units='kg*m**2')
            sharedIndeps.add_output('rna_cg', np.zeros(3), units='m')
            self.add_subsystem('sharedIndeps', sharedIndeps, promotes=['*'])

        # Define all input variables from all models
        floatingIndeps = IndepVarComp()
        
        # SemiGeometry
        floatingIndeps.add_output('radius_to_offset_column', 0.0, units='m')
        floatingIndeps.add_output('number_of_offset_columns', 0)
        
        floatingIndeps.add_output('fairlead_location', 0.0)
        floatingIndeps.add_output('fairlead_offset_from_shell', 0.0, units='m')
        floatingIndeps.add_output('z_offset', 0.0, units='m')
        floatingIndeps.add_output('max_draft', 0.0, units='m')

        # Mooring
        floatingIndeps.add_output('mooring_line_length', 0.0, units='m')
        floatingIndeps.add_output('anchor_radius', 0.0, units='m')
        floatingIndeps.add_output('mooring_diameter', 0.0, units='m')
        floatingIndeps.add_output('number_of_mooring_connections', 0)
        floatingIndeps.add_output('mooring_lines_per_connection', 0)
        floatingIndeps.add_discrete_output('mooring_type', 'chain')
        floatingIndeps.add_discrete_output('anchor_type', 'SUCTIONPILE')
        floatingIndeps.add_output('max_offset', 0.0, units='m')
        floatingIndeps.add_output('operational_heel', 0.0, units='deg')
        floatingIndeps.add_output('mooring_cost_factor', 0.0)
        floatingIndeps.add_output('max_survival_heel', 0.0, units='deg')

        # Column
        floatingIndeps.add_output('permanent_ballast_density', 0.0, units='kg/m**3')

        floatingIndeps.add_output('bulkhead_mass_factor', 0.0)
        floatingIndeps.add_output('ring_mass_factor', 0.0)
        floatingIndeps.add_output('shell_mass_factor', 0.0)
        floatingIndeps.add_output('column_mass_factor', 0.0)
        floatingIndeps.add_output('outfitting_mass_fraction', 0.0)
        floatingIndeps.add_output('ballast_cost_rate', 0.0, units='USD/kg')
        floatingIndeps.add_output('material_cost_rate', 0.0, units='USD/kg')
        floatingIndeps.add_output('labor_cost_rate', 0.0, units='USD/min')
        floatingIndeps.add_output('painting_cost_rate', 0.0, units='USD/m**2')
        floatingIndeps.add_output('outfitting_cost_rate', 0.0, units='USD/kg')
        floatingIndeps.add_discrete_output('loading', 'hydrostatic')
        
        floatingIndeps.add_output('max_taper_ratio', 0.0)
        floatingIndeps.add_output('min_diameter_thickness_ratio', 0.0)

        # Other Constraints
        floatingIndeps.add_output('wave_period_range_low', 2.0, units='s')
        floatingIndeps.add_output('wave_period_range_high', 20.0, units='s')


        self.add_subsystem('floatingIndeps', floatingIndeps, promotes=['*'])
        
        # Connect all input variables from all models
        self.connect('main.freeboard', 'tow.foundation_height')
        self.connect('z_offset', 'foundation_height')

        self.connect('tow.d_full', ['windLoads.d','tower_d_full']) # includes tower_d_full
        self.connect('tow.t_full', 'tower_t_full')
        self.connect('tow.z_full', ['loadingWind.z','windLoads.z','tower_z_full']) # includes tower_z_full
        self.connect('tow.cm.mass','tower_mass_section')
        self.connect('tow.turbine_mass','main.stack_mass_in')
        self.connect('tow.tower_center_of_mass','tower_center_of_mass')
        self.connect('tow.tower_raw_cost','tower_shell_cost')
        
        self.connect('max_taper_ratio', 'max_taper')
        self.connect('min_diameter_thickness_ratio', 'min_d_to_t')
        
        # To do: connect these to independent variables
        if topLevelFlag:
            self.connect('air_density',['main.windLoads.rho','off.windLoads.rho','windLoads.rho'])
            self.connect('air_viscosity',['main.windLoads.mu','off.windLoads.mu','windLoads.mu'])
            self.connect('water_density',['main.wave.rho','main.waveLoads.rho','off.wave.rho','off.waveLoads.rho'])
            self.connect('water_viscosity',['main.waveLoads.mu','off.waveLoads.mu'])
            self.connect('water_depth',['main.wave.z_floor','off.wave.z_floor'])
            self.connect('wave_beta',['main.waveLoads.beta','off.waveLoads.beta'])
            self.connect('wave_z0',['main.wave.z_surface','off.wave.z_surface'])
            self.connect('wind_z0','z0')
            self.connect('significant_wave_height',['Hs', 'hmax'])
            self.connect('significant_wave_period','T')
            self.connect('wind_reference_height','zref')
            self.connect('wind_reference_speed','Uref')
            
        
        self.connect('main.L_stiffener','main_buckling_length')
        self.connect('off.L_stiffener','offset_buckling_length')
        
        self.connect('bulkhead_mass_factor', 'buoyancy_tank_mass_factor')
        self.connect('shell_mass_factor', ['main.cyl_mass.outfitting_factor', 'off.cyl_mass.outfitting_factor'])

        self.connect('main.z_full', ['main_z_nodes', 'main_z_full'])
        self.connect('main.d_full', 'main_d_full')
        self.connect('main.t_full', 'main_t_full')

        self.connect('off.z_full', ['offset_z_nodes', 'offset_z_full'])
        self.connect('off.d_full', 'offset_d_full')
        self.connect('off.t_full', 'offset_t_full')

        self.connect('max_offset_restoring_force', 'mooring_surge_restoring_force')
        self.connect('operational_heel_restoring_force', 'mooring_pitch_restoring_force')
        
        self.connect('main.z_center_of_mass', 'main_center_of_mass')
        self.connect('main.z_center_of_buoyancy', 'main_center_of_buoyancy')
        self.connect('main.I_column', 'main_moments_of_inertia')
        self.connect('main.Iwater', 'main_Iwaterplane')
        self.connect('main.Awater', 'main_Awaterplane')
        self.connect('main.displaced_volume', 'main_displaced_volume')
        self.connect('main.hydrostatic_force', 'main_hydrostatic_force')
        self.connect('main.column_added_mass', 'main_added_mass')
        self.connect('main.column_total_mass', 'main_mass')
        self.connect('main.column_total_cost', 'main_cost')
        self.connect('main.variable_ballast_interp_zpts', 'water_ballast_zpts_vector')
        self.connect('main.variable_ballast_interp_radius', 'water_ballast_radius_vector')
        self.connect('main.Px', 'main_Px')
        self.connect('main.Py', 'main_Py')
        self.connect('main.Pz', 'main_Pz')
        self.connect('main.qdyn', 'main_qdyn')

        self.connect('off.z_center_of_mass', 'offset_center_of_mass')
        self.connect('off.z_center_of_buoyancy', 'offset_center_of_buoyancy')
        self.connect('off.I_column', 'offset_moments_of_inertia')
        self.connect('off.Iwater', 'offset_Iwaterplane')
        self.connect('off.Awater', 'offset_Awaterplane')
        self.connect('off.displaced_volume', 'offset_displaced_volume')
        self.connect('off.hydrostatic_force', 'offset_hydrostatic_force')
        self.connect('off.column_added_mass', 'offset_added_mass')
        self.connect('off.column_total_mass', 'offset_mass')
        self.connect('off.column_total_cost', 'offset_cost')
        self.connect('off.Px', 'offset_Px')
        self.connect('off.Py', 'offset_Py')
        self.connect('off.Pz', 'offset_Pz')
        self.connect('off.qdyn', 'offset_qdyn')
        self.connect('off.draft', 'offset_draft')




def commonVars(prob, nsection):
    # Variables common to both examples

    # Set environment to that used in OC4 testing campaign
    prob['shearExp']    = 0.11   # Shear exponent in wind power law
    prob['cm']          = 2.0    # Added mass coefficient
    prob['Uc']          = 0.0    # Mean current speed
    prob['wind_z0']     = 0.0    # Water line
    prob['yaw']         = 0.0    # Turbine yaw angle
    prob['beta']        = 0.0    # Wind beta angle
    prob['cd_usr']      = -1.0 # Compute drag coefficient

    # Wind and water properties
    prob['air_density'] = 1.226   # Density of air [kg/m^3]
    prob['air_viscosity']  = 1.78e-5 # Viscosity of air [kg/m/s]
    prob['water_density']      = 1025.0  # Density of water [kg/m^3]
    prob['water_viscosity']  = 1.08e-3 # Viscosity of water [kg/m/s]
    
    # Material properties
    prob['material_density'] = prob['main.material_density'] = prob['off.material_density'] = 7850.0          # Steel [kg/m^3]
    prob['E']                = 200e9           # Young's modulus [N/m^2]
    prob['G']                = 79.3e9          # Shear modulus [N/m^2]
    prob['yield_stress']     = 3.45e8          # Elastic yield stress [N/m^2]
    prob['nu']               = 0.26            # Poisson's ratio
    prob['permanent_ballast_density'] = 4492.0 # [kg/m^3]

    # Mass and cost scaling factors
    prob['bulkhead_mass_factor']     = 1.0     # Scaling for unaccounted bulkhead mass
    prob['ring_mass_factor']         = 1.0     # Scaling for unaccounted stiffener mass
    prob['shell_mass_factor']        = 1.0     # Scaling for unaccounted shell mass
    prob['column_mass_factor']       = 1.05    # Scaling for unaccounted column mass
    prob['outfitting_mass_fraction'] = 0.06    # Fraction of additional outfitting mass for each column
    prob['ballast_cost_rate']        = 0.1   # Cost factor for ballast mass [$/kg]
    prob['material_cost_rate']       = 1.1  # Cost factor for column mass [$/kg]
    prob['labor_cost_rate']          = 1.0  # Cost factor for labor time [$/min]
    prob['painting_cost_rate']       = 14.4  # Cost factor for column surface finishing [$/m^2]
    prob['outfitting_cost_rate']     = 1.5*1.1  # Cost factor for outfitting mass [$/kg]
    prob['mooring_cost_factor']      = 1.1     # Cost factor for mooring mass [$/kg]
    
    # Safety factors
    prob['gamma_f'] = 1.35 # Safety factor on loads
    prob['gamma_b'] = 1.1  # Safety factor on buckling
    prob['gamma_m'] = 1.1  # Safety factor on materials
    prob['gamma_n'] = 1.0  # Safety factor on consequence of failure
    prob['gamma_fatigue'] = 1.755 # Not used

    # Mooring parameters
    prob['number_of_mooring_connections'] = 3             # Evenly spaced around structure
    prob['mooring_lines_per_connection'] = 1             # Evenly spaced around structure
    prob['mooring_type']               = 'chain'       # Options are chain, nylon, polyester, fiber, or iwrc
    prob['anchor_type']                = 'DRAGEMBEDMENT' # Options are SUCTIONPILE or DRAGEMBEDMENT
    
    # Porperties of turbine tower
    nTower = prob.model.options['nTower']
    prob['hub_height']              = 77.6                              # Length from tower main to top (not including freeboard) [m]
    prob['tower_section_height']    = 77.6/nTower * np.ones(nTower) # Length of each tower section [m]
    prob['tower_outer_diameter']    = np.linspace(6.5, 3.87, nTower+1) # Diameter at each tower section node (linear lofting between) [m]
    prob['tower_wall_thickness']    = np.linspace(0.027, 0.019, nTower) # Diameter at each tower section node (linear lofting between) [m]
    prob['tower_buckling_length']   = 30.0                              # Tower buckling reinforcement spacing [m]
    prob['tower_outfitting_factor'] = 1.07                              # Scaling for unaccounted tower mass in outfitting

    # Properties of rotor-nacelle-assembly (RNA)
    prob['rna_mass']   = 350e3 # Mass [kg]
    prob['rna_I']      = 1e5*np.array([1149.307, 220.354, 187.597, 0, 5.037, 0]) # Moment of intertia (xx,yy,zz,xy,xz,yz) [kg/m^2]
    prob['rna_cg']     = np.array([-1.132, 0, 0.509])                       # Offset of RNA center of mass from tower top (x,y,z) [m]
    # Max thrust
    prob['rna_force']  = np.array([1284744.196, 0, -112400.5527])           # Net force acting on RNA (x,y,z) [N]
    prob['rna_moment'] = np.array([3963732.762, 896380.8464, -346781.682]) # Net moment acting on RNA (x,y,z) [N*m]
    # Max wind speed
    #prob['rna_force']  = np.array([188038.8045, 0,  -16451.2637]) # Net force acting on RNA (x,y,z) [N]
    #prob['rna_moment'] = np.array([0.0, 131196.8431,  0.0]) # Net moment acting on RNA (x,y,z) [N*m]
    
    # Mooring constraints
    prob['max_draft'] = 150.0 # Max surge/sway offset [m]      
    prob['max_offset'] = 100.0 # Max surge/sway offset [m]      
    prob['operational_heel']   = 10.0 # Max heel (pitching) angle [deg]

    # Design constraints
    prob['max_taper_ratio'] = 0.2                # For manufacturability of rolling steel
    prob['min_diameter_thickness_ratio'] = 120.0 # For weld-ability
    prob['connection_ratio_max']      = 0.25 # For welding pontoons to columns

    # API 2U flag
    prob['loading'] = 'hydrostatic'
    
    return prob


def sparExample():
    # Initialize OpenMDAO problem and FloatingSE Group
    prob = Problem()
    prob.model = FloatingSE()
    prob.setup()

    # Variables common to both examples
    nsection = prob.model.options['nSection']
    prob = commonVars(prob, nsection)
    
    # Remove all offset columns
    prob['number_of_offset_columns'] = 0
    prob['cross_attachment_pontoons_int']   = 0
    prob['lower_attachment_pontoons_int']   = 0
    prob['upper_attachment_pontoons_int']   = 0
    prob['lower_ring_pontoons_int']         = 0
    prob['upper_ring_pontoons_int']         = 0
    prob['outer_cross_pontoons_int']        = 0

    # Set environment to that used in OC3 testing campaign
    prob['water_depth'] = 320.0  # Distance to sea floor [m]
    prob['significant_wave_height']        = 10.8   # Significant wave height [m]
    prob['significant_wave_period']           = 9.8    # Wave period [s]
    prob['wind_reference_speed']        = 11.0   # Wind reference speed [m/s]
    prob['wind_reference_height']        = 119.0  # Wind reference height [m]

    # Column geometry
    prob['main.permanent_ballast_height'] = 10.0 # Height above keel for permanent ballast [m]
    prob['main.freeboard']                = 10.0 # Height extension above waterline [m]
    prob['main.section_height'] = np.array([49.0, 59.0, 8.0, 14.0])  # Length of each section [m]
    prob['main.outer_diameter'] = np.array([9.4, 9.4, 9.4, 6.5, 6.5]) # Diameter at each section node (linear lofting between) [m]
    prob['main.wall_thickness'] = 0.05 * np.ones(nsection)               # Shell thickness at each section node (linear lofting between) [m]
    prob['main.bulkhead_thickness'] = 0.05*np.array([1, 1, 0, 1, 0]) # Locations/thickness of internal bulkheads at section interfaces [m]
    
    # Column ring stiffener parameters
    prob['main.stiffener_web_height']       = 0.10 * np.ones(nsection) # (by section) [m]
    prob['main.stiffener_web_thickness']    = 0.04 * np.ones(nsection) # (by section) [m]
    prob['main.stiffener_flange_width']     = 0.10 * np.ones(nsection) # (by section) [m]
    prob['main.stiffener_flange_thickness'] = 0.02 * np.ones(nsection) # (by section) [m]
    prob['main.stiffener_spacing']          = 0.40 * np.ones(nsection) # (by section) [m]
    
    # Mooring parameters
    prob['mooring_diameter']           = 0.09          # Diameter of mooring line/chain [m]
    prob['fairlead']                   = 70.0          # Distance below waterline for attachment [m]
    prob['fairlead_offset_from_shell'] = 0.5           # Offset from shell surface for mooring attachment [m]
    prob['mooring_line_length']        = 902.2         # Unstretched mooring line length
    prob['anchor_radius']              = 853.87        # Distance from centerline to sea floor landing [m]
    prob['fairlead_support_outer_diameter'] = 3.2    # Diameter of all fairlead support elements [m]
    prob['fairlead_support_wall_thickness'] = 0.0175 # Thickness of all fairlead support elements [m]

    # Other variables to avoid divide by zeros, even though it won't matter
    prob['radius_to_offset_column'] = 15.0
    prob['off.freeboard'] = 0.1
    prob['off.section_height'] = 1.0 * np.ones(nsection)
    prob['off.outer_diameter'] = 5.0 * np.ones(nsection+1)
    prob['off.wall_thickness'] = 0.1 * np.ones(nsection)
    prob['off.permanent_ballast_height'] = 0.1
    prob['off.stiffener_web_height'] = 0.1 * np.ones(nsection)
    prob['off.stiffener_web_thickness'] =  0.1 * np.ones(nsection)
    prob['off.stiffener_flange_width'] =  0.1 * np.ones(nsection)
    prob['off.stiffener_flange_thickness'] =  0.1 * np.ones(nsection)
    prob['off.stiffener_spacing'] =  0.1 * np.ones(nsection)
    prob['pontoon_outer_diameter'] = 1.0
    prob['pontoon_wall_thickness'] = 0.1
    
    prob.run_model()
    


def semiExample():
    # Initialize OpenMDAO problem and FloatingSE Group
    prob = Problem()
    prob.model = FloatingSE()
    prob.setup()

    # Variables common to both examples
    nsection = prob.model.options['nSection']
    prob = commonVars(prob, nsection)
    
    # Add in offset columns and truss elements
    prob['number_of_offset_columns'] = 3
    prob['cross_attachment_pontoons_int']   = 1 # Lower-Upper main-to-offset connecting cross braces
    prob['lower_attachment_pontoons_int']   = 1 # Lower main-to-offset connecting pontoons
    prob['upper_attachment_pontoons_int']   = 1 # Upper main-to-offset connecting pontoons
    prob['lower_ring_pontoons_int']         = 1 # Lower ring of pontoons connecting offset columns
    prob['upper_ring_pontoons_int']         = 1 # Upper ring of pontoons connecting offset columns
    prob['outer_cross_pontoons_int']        = 1 # Auxiliary ring connecting V-cross braces

    # Set environment to that used in OC4 testing campaign
    prob['water_depth'] = 200.0  # Distance to sea floor [m]
    prob['significant_wave_height']        = 10.8   # Significant wave height [m]
    prob['significant_wave_period']           = 9.8    # Wave period [s]
    prob['wind_reference_speed']        = 11.0   # Wind reference speed [m/s]
    prob['wind_reference_height']        = 119.0  # Wind reference height [m]

    # Column geometry
    prob['main.permanent_ballast_height'] = 10.0 # Height above keel for permanent ballast [m]
    prob['main.freeboard']                = 10.0 # Height extension above waterline [m]
    prob['main.section_height'] = np.array([49.0, 59.0, 8.0, 14.0])  # Length of each section [m]
    prob['main.outer_diameter'] = np.array([9.4, 9.4, 9.4, 6.5, 6.5]) # Diameter at each section node (linear lofting between) [m]
    prob['main.wall_thickness'] = 0.05 * np.ones(nsection)               # Shell thickness at each section node (linear lofting between) [m]
    prob['main.bulkhead_thickness'] = 0.05*np.array([1, 1, 0, 1, 0]) # Locations/thickness of internal bulkheads at section interfaces [m]

    # Auxiliary column geometry
    prob['radius_to_offset_column']         = 33.333 * np.cos(np.pi/6) # Centerline of main column to centerline of offset column [m]
    prob['off.permanent_ballast_height'] = 0.1                      # Height above keel for permanent ballast [m]
    prob['off.freeboard']                = 12.0                     # Height extension above waterline [m]
    prob['off.section_height']           = np.array([6.0, 0.1, 15.9, 10]) # Length of each section [m]
    prob['off.outer_diameter']           = np.array([24, 24, 12, 12, 12]) # Diameter at each section node (linear lofting between) [m]
    prob['off.wall_thickness']           = 0.06 * np.ones(nsection)         # Shell thickness at each section node (linear lofting between) [m]

    # Column ring stiffener parameters
    prob['main.stiffener_web_height']       = 0.10 * np.ones(nsection) # (by section) [m]
    prob['main.stiffener_web_thickness']    = 0.04 * np.ones(nsection) # (by section) [m]
    prob['main.stiffener_flange_width']     = 0.10 * np.ones(nsection) # (by section) [m]
    prob['main.stiffener_flange_thickness'] = 0.02 * np.ones(nsection) # (by section) [m]
    prob['main.stiffener_spacing']          = 0.40 * np.ones(nsection) # (by section) [m]

    # Auxiliary column ring stiffener parameters
    prob['off.stiffener_web_height']       = 0.10 * np.ones(nsection) # (by section) [m]
    prob['off.stiffener_web_thickness']    = 0.04 * np.ones(nsection) # (by section) [m]
    prob['off.stiffener_flange_width']     = 0.01 * np.ones(nsection) # (by section) [m]
    prob['off.stiffener_flange_thickness'] = 0.02 * np.ones(nsection) # (by section) [m]
    prob['off.stiffener_spacing']          = 0.40 * np.ones(nsection) # (by section) [m]

    # Pontoon parameters
    prob['pontoon_outer_diameter']    = 3.2    # Diameter of all pontoon/truss elements [m]
    prob['pontoon_wall_thickness']    = 0.0175 # Thickness of all pontoon/truss elements [m]
    prob['main_pontoon_attach_lower'] = -20.0  # Lower z-coordinate on main where truss attaches [m]
    prob['main_pontoon_attach_upper'] = 10.0   # Upper z-coordinate on main where truss attaches [m]
    
    # Mooring parameters
    prob['mooring_diameter']           = 0.0766        # Diameter of mooring line/chain [m]
    prob['fairlead']                   = 14.0          # Distance below waterline for attachment [m]
    prob['fairlead_offset_from_shell'] = 0.5           # Offset from shell surface for mooring attachment [m]
    prob['mooring_line_length']        = 835.5+300         # Unstretched mooring line length
    prob['anchor_radius']              = 837.6+300.0         # Distance from centerline to sea floor landing [m]
    prob['fairlead_support_outer_diameter'] = 3.2    # Diameter of all fairlead support elements [m]
    prob['fairlead_support_wall_thickness'] = 0.0175 # Thickness of all fairlead support elements [m]
    
    prob.run_model()
    
    '''
    f = open('deriv_semi.dat','w')
    out = prob.check_total_derivatives(f)
    #out = prob.check_partial_derivatives(f, compact_print=True)
    f.close()
    tol = 1e-4
    for comp in out.keys():
        for k in out[comp].keys():
            if ( (out[comp][k]['rel error'][0] > tol) and (out[comp][k]['abs error'][0] > tol) ):
                print k
    '''

if __name__ == "__main__":
    from openmdao.api import Problem
    import sys

    if len(sys.argv) > 1 and sys.argv[1].lower() in ['spar','column','col','oc3']:
        sparExample()
    else:
        semiExample()
        




import numpy as np
import os

from rotorse.precomp import Profile, Orthotropic2DMaterial, CompositeSection  # TODO: can just pass file names and do this initialization inside of rotor
#from towerse.tower import TowerWithpBEAM
from commonse.environment import PowerWind, TowerSoil, LinearWaves
from jacketse.jacket import JcktGeoInputs,SoilGeoInputs,WaterInputs,WindInputs,RNAprops,TPlumpMass,Frame3DDaux,\
                    MatInputs,LegGeoInputs,XBrcGeoInputs,MudBrcGeoInputs,HBrcGeoInputs,TPGeoInputs,PileGeoInputs,\
                    TwrGeoInputs
from commonse.utilities import cosd, sind
from rotorse.rotoraero import RS2RPM


def configure_nrel5mw_turbine_with_jacket(turbine,wind_class='I',sea_depth = 0.0):
    """
    Inputs:
        rotor = RotorSE()
        nacelle = DriveSE()
        jacket = JacketSE()
        wind_class : str ('I', 'III', 'Offshore' - selected wind class for project)
        sea_depth : float (sea depth if an offshore wind plant)
    """

    # =================


    # === Turbine ===
    turbine.rho = 1.225  # (Float, kg/m**3): density of air
    turbine.mu = 1.81206e-5  # (Float, kg/m/s): dynamic viscosity of air
    turbine.shear_exponent = 0.2  # (Float): shear exponent
    turbine.hub_height = 90.0  # (Float, m): hub height
    turbine.turbine_class = 'I'  # (Enum): IEC turbine class
    turbine.turbulence_class = 'B'  # (Enum): IEC turbulence class class
    turbine.cdf_reference_height_wind_speed = 90.0  # (Float): reference hub height for IEC wind speed (used in CDF calculation)
    turbine.g = 9.81  # (Float, m/s**2): acceleration of gravity
    # ======================

    # === rotor ===
    # --- blade grid ---
    turbine.rotor.initial_aero_grid = np.array([0.02222276, 0.06666667, 0.11111057, 0.16666667, 0.23333333, 0.3, 0.36666667,
        0.43333333, 0.5, 0.56666667, 0.63333333, 0.7, 0.76666667, 0.83333333, 0.88888943, 0.93333333,
        0.97777724])  # (Array): initial aerodynamic grid on unit radius
    turbine.rotor.initial_str_grid = np.array([0.0, 0.00492790457512, 0.00652942887106, 0.00813095316699, 0.00983257273154,
        0.0114340970275, 0.0130356213234, 0.02222276, 0.024446481932, 0.026048006228, 0.06666667, 0.089508406455,
        0.11111057, 0.146462614229, 0.16666667, 0.195309105255, 0.23333333, 0.276686558545, 0.3, 0.333640766319,
        0.36666667, 0.400404310407, 0.43333333, 0.5, 0.520818918408, 0.56666667, 0.602196371696, 0.63333333,
        0.667358391486, 0.683573824984, 0.7, 0.73242031601, 0.76666667, 0.83333333, 0.88888943, 0.93333333, 0.97777724,
        1.0])  # (Array): initial structural grid on unit radius
    turbine.rotor.idx_cylinder_aero = 3  # (Int): first idx in r_aero_unit of non-cylindrical section, constant twist inboard of here
    turbine.rotor.idx_cylinder_str = 14  # (Int): first idx in r_str_unit of non-cylindrical section
    turbine.rotor.hubFraction = 0.025  # (Float): hub location as fraction of radius
    # ------------------

    # --- blade geometry ---
    turbine.rotor.r_aero = np.array([0.02222276, 0.06666667, 0.11111057, 0.2, 0.23333333, 0.3, 0.36666667, 0.43333333,
        0.5, 0.56666667, 0.63333333, 0.64, 0.7, 0.83333333, 0.88888943, 0.93333333,
        0.97777724])  # (Array): new aerodynamic grid on unit radius
    turbine.rotor.r_max_chord = 0.23577  # (Float): location of max chord on unit radius
    turbine.rotor.chord_sub = [3.2612, 4.5709, 3.3178, 1.4621]  # (Array, m): chord at control points. defined at hub, then at linearly spaced locations from r_max_chord to tip
    turbine.rotor.theta_sub = [13.2783, 7.46036, 2.89317, -0.0878099]  # (Array, deg): twist at control points.  defined at linearly spaced locations from r[idx_cylinder] to tip
    turbine.rotor.precurve_sub = [0.0, 0.0, 0.0]  # (Array, m): precurve at control points.  defined at same locations at chord, starting at 2nd control point (root must be zero precurve)
    turbine.rotor.delta_precurve_sub = [0.0, 0.0, 0.0]  # (Array, m): adjustment to precurve to account for curvature from loading
    turbine.rotor.sparT = [0.05, 0.047754, 0.045376, 0.031085, 0.0061398]  # (Array, m): spar cap thickness parameters
    turbine.rotor.teT = [0.1, 0.09569, 0.06569, 0.02569, 0.00569]  # (Array, m): trailing-edge thickness parameters
    turbine.rotor.bladeLength = 61.5  # (Float, m): blade length (if not precurved or swept) otherwise length of blade before curvature
    turbine.rotor.delta_bladeLength = 0.0  # (Float, m): adjustment to blade length to account for curvature from loading
    turbine.rotor.precone = 2.5  # (Float, deg): precone angle
    turbine.rotor.tilt = 5.0  # (Float, deg): shaft tilt
    turbine.rotor.yaw = 0.0  # (Float, deg): yaw error
    turbine.rotor.nBlades = 3  # (Int): number of blades
    # ------------------

    # --- airfoil files ---
    import rotorse
    #basepath = os.path.join('5MW_files', '5MW_AFFiles')
    basepath = os.path.join('..','reference_turbines','nrel5mw','airfoils')

    # load all airfoils
    airfoil_types = [0]*8
    airfoil_types[0] = os.path.join(basepath, 'Cylinder1.dat')
    airfoil_types[1] = os.path.join(basepath, 'Cylinder2.dat')
    airfoil_types[2] = os.path.join(basepath, 'DU40_A17.dat')
    airfoil_types[3] = os.path.join(basepath, 'DU35_A17.dat')
    airfoil_types[4] = os.path.join(basepath, 'DU30_A17.dat')
    airfoil_types[5] = os.path.join(basepath, 'DU25_A17.dat')
    airfoil_types[6] = os.path.join(basepath, 'DU21_A17.dat')
    airfoil_types[7] = os.path.join(basepath, 'NACA64_A17.dat')

    # place at appropriate radial stations
    af_idx = [0, 0, 1, 2, 3, 3, 4, 5, 5, 6, 6, 7, 7, 7, 7, 7, 7]

    n = len(af_idx)
    af = [0]*n
    for i in range(n):
        af[i] = airfoil_types[af_idx[i]]
    turbine.rotor.airfoil_files = af  # (List): names of airfoil file
    # ----------------------

    # --- control ---
    turbine.rotor.control.Vin = 3.0  # (Float, m/s): cut-in wind speed
    turbine.rotor.control.Vout = 25.0  # (Float, m/s): cut-out wind speed
    turbine.rotor.control.ratedPower = 5e6  # (Float, W): rated power
    turbine.rotor.control.minOmega = 0.0  # (Float, rpm): minimum allowed rotor rotation speed
    turbine.rotor.control.maxOmega = 12.0  # (Float, rpm): maximum allowed rotor rotation speed
    turbine.rotor.control.tsr = 7.55  # (Float): tip-speed ratio in Region 2 (should be optimized externally)
    turbine.rotor.control.pitch = 0.0  # (Float, deg): pitch angle in region 2 (and region 3 for fixed pitch machines)
    turbine.rotor.pitch_extreme = 0.0  # (Float, deg): worst-case pitch at survival wind condition
    turbine.rotor.azimuth_extreme = 0.0  # (Float, deg): worst-case azimuth at survival wind condition
    turbine.rotor.VfactorPC = 0.7  # (Float): fraction of rated speed at which the deflection is assumed to representative throughout the power curve calculation
    # ----------------------

    # --- aero and structural analysis options ---
    turbine.rotor.nSector = 4  # (Int): number of sectors to divide rotor face into in computing thrust and power
    turbine.rotor.npts_coarse_power_curve = 20  # (Int): number of points to evaluate aero analysis at
    turbine.rotor.npts_spline_power_curve = 200  # (Int): number of points to use in fitting spline to power curve
    turbine.rotor.AEP_loss_factor = 1.0  # (Float): availability and other losses (soiling, array, etc.)
    turbine.rotor.drivetrainType = 'geared'  # (Enum)
    turbine.rotor.nF = 5  # (Int): number of natural frequencies to compute
    turbine.rotor.dynamic_amplication_tip_deflection = 1.35  # (Float): a dynamic amplification factor to adjust the static deflection calculation
    # ----------------------


    # --- materials and composite layup  ---
    #basepath = os.path.join('5MW_files', '5MW_PrecompFiles')
    basepath = os.path.join('..', 'reference_turbines','nrel5mw','blade')

    materials = Orthotropic2DMaterial.listFromPreCompFile(os.path.join(basepath, 'materials.inp'))

    ncomp = len(turbine.rotor.initial_str_grid)
    upper = [0]*ncomp
    lower = [0]*ncomp
    webs = [0]*ncomp
    profile = [0]*ncomp

    turbine.rotor.leLoc = np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.498, 0.497, 0.465, 0.447, 0.43, 0.411,
        0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4, 0.4,
        0.4, 0.4, 0.4, 0.4])    # (Array): array of leading-edge positions from a reference blade axis (usually blade pitch axis). locations are normalized by the local chord length. e.g. leLoc[i] = 0.2 means leading edge is 0.2*chord[i] from reference axis.  positive in -x direction for airfoil-aligned coordinate system
    turbine.rotor.sector_idx_strain_spar = [2]*ncomp  # (Array): index of sector for spar (PreComp definition of sector)
    turbine.rotor.sector_idx_strain_te = [3]*ncomp  # (Array): index of sector for trailing-edge (PreComp definition of sector)
    web1 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, 0.4114, 0.4102, 0.4094, 0.3876, 0.3755, 0.3639, 0.345, 0.3342, 0.3313, 0.3274, 0.323, 0.3206, 0.3172, 0.3138, 0.3104, 0.307, 0.3003, 0.2982, 0.2935, 0.2899, 0.2867, 0.2833, 0.2817, 0.2799, 0.2767, 0.2731, 0.2664, 0.2607, 0.2562, 0.1886, -1.0])
    web2 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, 0.5886, 0.5868, 0.5854, 0.5508, 0.5315, 0.5131, 0.4831, 0.4658, 0.4687, 0.4726, 0.477, 0.4794, 0.4828, 0.4862, 0.4896, 0.493, 0.4997, 0.5018, 0.5065, 0.5101, 0.5133, 0.5167, 0.5183, 0.5201, 0.5233, 0.5269, 0.5336, 0.5393, 0.5438, 0.6114, -1.0])
    web3 = np.array([-1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0, -1.0])
    turbine.rotor.chord_str_ref = np.array([3.2612, 3.3100915356, 3.32587052924, 3.34159388653, 3.35823798667, 3.37384375335,
        3.38939112914, 3.4774055542, 3.49839685, 3.51343645709, 3.87017220335, 4.04645623801, 4.19408216643,
         4.47641008477, 4.55844487985, 4.57383098262, 4.57285771934, 4.51914315648, 4.47677655262, 4.40075650022,
         4.31069949379, 4.20483735936, 4.08985563932, 3.82931757126, 3.74220276467, 3.54415796922, 3.38732428502,
         3.24931446473, 3.23421422609, 3.22701537997, 3.21972125648, 3.08979310611, 2.95152261813, 2.330753331,
         2.05553464181, 1.82577817774, 1.5860853279, 1.4621])  # (Array, m): chord distribution for reference section, thickness of structural layup scaled with reference thickness (fixed t/c for this case)

    for i in range(ncomp):

        webLoc = []
        if web1[i] != -1:
            webLoc.append(web1[i])
        if web2[i] != -1:
            webLoc.append(web2[i])
        if web3[i] != -1:
            webLoc.append(web3[i])

        upper[i], lower[i], webs[i] = CompositeSection.initFromPreCompLayupFile(os.path.join(basepath, 'layup_' + str(i+1) + '.inp'), webLoc, materials)
        profile[i] = Profile.initFromPreCompFile(os.path.join(basepath, 'shape_' + str(i+1) + '.inp'))

    turbine.rotor.materials = materials  # (List): list of all Orthotropic2DMaterial objects used in defining the geometry
    turbine.rotor.upperCS = upper  # (List): list of CompositeSection objections defining the properties for upper surface
    turbine.rotor.lowerCS = lower  # (List): list of CompositeSection objections defining the properties for lower surface
    turbine.rotor.websCS = webs  # (List): list of CompositeSection objections defining the properties for shear webs
    turbine.rotor.profile = profile  # (List): airfoil shape at each radial position
    # --------------------------------------

    strain_ult_spar = 1.0e-2
    strain_ult_te = 2500*1e-6

    # --- fatigue ---
    turbine.rotor.rstar_damage = np.array([0.000, 0.022, 0.067, 0.111, 0.167, 0.233, 0.300, 0.367, 0.433, 0.500,
        0.567, 0.633, 0.700, 0.767, 0.833, 0.889, 0.933, 0.978])  # (Array): nondimensional radial locations of damage equivalent moments
    turbine.rotor.Mxb_damage = 1e3*np.array([2.3743E+003, 2.0834E+003, 1.8108E+003, 1.5705E+003, 1.3104E+003,
        1.0488E+003, 8.2367E+002, 6.3407E+002, 4.7727E+002, 3.4804E+002, 2.4458E+002, 1.6339E+002,
        1.0252E+002, 5.7842E+001, 2.7349E+001, 1.1262E+001, 3.8549E+000, 4.4738E-001])  # (Array, N*m): damage equivalent moments about blade c.s. x-direction
    turbine.rotor.Myb_damage = 1e3*np.array([2.7732E+003, 2.8155E+003, 2.6004E+003, 2.3933E+003, 2.1371E+003,
        1.8459E+003, 1.5582E+003, 1.2896E+003, 1.0427E+003, 8.2015E+002, 6.2449E+002, 4.5229E+002,
        3.0658E+002, 1.8746E+002, 9.6475E+001, 4.2677E+001, 1.5409E+001, 1.8426E+000])  # (Array, N*m): damage equivalent moments about blade c.s. y-direction
    turbine.rotor.strain_ult_spar = 1.0e-2  # (Float): ultimate strain in spar cap
    turbine.rotor.strain_ult_te = 2500*1e-6 * 2   # (Float): uptimate strain in trailing-edge panels, note that I am putting a factor of two for the damage part only.
    turbine.rotor.eta_damage = 1.35*1.3*1.0  # (Float): safety factor for fatigue
    turbine.rotor.m_damage = 10.0  # (Float): slope of S-N curve for fatigue analysis
    turbine.rotor.N_damage = 365*24*3600*20.0  # (Float): number of cycles used in fatigue analysis  TODO: make function of rotation speed
    # ----------------
    # =================


    # === nacelle ======
    turbine.nacelle.L_ms = 1.0  # (Float, m): main shaft length downwind of main bearing in low-speed shaft
    turbine.nacelle.L_mb = 2.5  # (Float, m): main shaft length in low-speed shaft

    turbine.nacelle.h0_front = 1.7  # (Float, m): height of Ibeam in bedplate front
    turbine.nacelle.h0_rear = 1.35  # (Float, m): height of Ibeam in bedplate rear

    turbine.nacelle.drivetrain_design = 'geared'
    turbine.nacelle.crane = True  # (Bool): flag for presence of crane
    turbine.nacelle.bevel = 0  # (Int): Flag for the presence of a bevel stage - 1 if present, 0 if not
    turbine.nacelle.gear_configuration = 'eep'  # (Str): tring that represents the configuration of the gearbox (stage number and types)

    turbine.nacelle.Np = [3, 3, 1]  # (Array): number of planets in each stage
    turbine.nacelle.ratio_type = 'optimal'  # (Str): optimal or empirical stage ratios
    turbine.nacelle.shaft_type = 'normal'  # (Str): normal or short shaft length
    #turbine.nacelle.shaft_angle = 5.0  # (Float, deg): Angle of the LSS inclindation with respect to the horizontal
    turbine.nacelle.shaft_ratio = 0.10  # (Float): Ratio of inner diameter to outer diameter.  Leave zero for solid LSS
    turbine.nacelle.carrier_mass = 8000.0 # estimated for 5 MW
    turbine.nacelle.mb1Type = 'CARB'  # (Str): Main bearing type: CARB, TRB or SRB
    turbine.nacelle.mb2Type = 'SRB'  # (Str): Second bearing type: CARB, TRB or SRB
    turbine.nacelle.yaw_motors_number = 8.0  # (Float): number of yaw motors
    turbine.nacelle.uptower_transformer = True
    turbine.nacelle.flange_length = 0.5 #m
    turbine.nacelle.gearbox_cm = 0.1
    turbine.nacelle.hss_length = 1.5
    turbine.nacelle.overhang = 5.0 #TODO - should come from turbine configuration level

    turbine.nacelle.check_fatigue = 0 #0 if no fatigue check, 1 if parameterized fatigue check, 2 if known loads inputs

    # TODO: should come from rotor (these are FAST outputs)
    turbine.nacelle.DrivetrainEfficiency = 0.95
    turbine.nacelle.rotor_bending_moment_x = 330770.0# Nm
    turbine.nacelle.rotor_bending_moment_y = -16665000.0 # Nm
    turbine.nacelle.rotor_bending_moment_z = 2896300.0 # Nm
    turbine.nacelle.rotor_force_x = 599610.0 # N
    turbine.nacelle.rotor_force_y = 186780.0 # N
    turbine.nacelle.rotor_force_z = -842710.0 # N

    #turbine.nacelle.h0_rear = 1.35 # only used in drive smooth
    #turbine.nacelle.h0_front = 1.7

    # =================

    if wind_class == 'I':
        turbine.rotor.turbine_class = 'I'

    elif wind_class == 'III':
        turbine.rotor.turbine_class = 'III'

        # for fatigue based analysis of class III wind turbine
        turbine.tower.M_DEL = 1.028713178 * 1e3*np.array([7.8792E+003, 7.7507E+003, 7.4918E+003, 7.2389E+003, 6.9815E+003, 6.7262E+003, 6.4730E+003, 6.2174E+003, 5.9615E+003, 5.7073E+003, 5.4591E+003, 5.2141E+003, 4.9741E+003, 4.7399E+003, 4.5117E+003, 4.2840E+003, 4.0606E+003, 3.8360E+003, 3.6118E+003, 3.3911E+003, 3.1723E+003, 2.9568E+003, 2.7391E+003, 2.5294E+003, 2.3229E+003, 2.1246E+003, 1.9321E+003, 1.7475E+003, 1.5790E+003, 1.4286E+003, 1.3101E+003, 1.2257E+003, 1.1787E+003, 1.1727E+003, 1.1821E+003])

        turbine.rotor.Mxb_damage = 1e3*np.array([2.3617E+003, 2.0751E+003, 1.8051E+003, 1.5631E+003, 1.2994E+003, 1.0388E+003, 8.1384E+002, 6.2492E+002, 4.6916E+002, 3.4078E+002, 2.3916E+002, 1.5916E+002, 9.9752E+001, 5.6139E+001, 2.6492E+001, 1.0886E+001, 3.7210E+000, 4.3206E-001])
        turbine.rotor.Myb_damage = 1e3*np.array([2.5492E+003, 2.6261E+003, 2.4265E+003, 2.2308E+003, 1.9882E+003, 1.7184E+003, 1.4438E+003, 1.1925E+003, 9.6251E+002, 7.5564E+002, 5.7332E+002, 4.1435E+002, 2.8036E+002, 1.7106E+002, 8.7732E+001, 3.8678E+001, 1.3942E+001, 1.6600E+000])

    elif wind_class == 'Offshore':
        turbine.rotor.turbine_class = 'I'

    # =================

    # === jacket ===

    #--- Set Jacket Input Parameters ---#
    Jcktins=JcktGeoInputs()
    Jcktins.nlegs =4
    Jcktins.nbays =5
    Jcktins.batter=12.
    Jcktins.dck_botz =16.
    Jcktins.weld2D   =0.5
    Jcktins.VPFlag = True    #vertical pile T/F;  to enable piles in frame3DD set pileinputs.ndiv>0
    Jcktins.clamped= False    #whether or not the bottom of the structure is rigidly connected. Use False when equivalent spring constants are being used.
    Jcktins.AFflag = False  #whether or not to use apparent fixity piles
    Jcktins.PreBuildTPLvl = 2  #if >0, the TP is prebuilt according to rules per PreBuildTP

    #Soil inputs
    Soilinputs=SoilGeoInputs()
    Soilinputs.zbots   =-np.array([3.,5.,7.,15.,30.,50.])
    Soilinputs.gammas  =np.array([10000.,10000.,10000.,10000.,10000.,10000.])
    Soilinputs.cus     =np.array([60000.,60000.,60000.,60000.,60000.,60000.])
    Soilinputs.phis    =np.array([26.,26.,26.,26.,26.,26])#np.array([36.,33.,26.,37.,35.,37.5])#np.array([36.,33.,26.,37.,35.,37.5])
    Soilinputs.delta   =25.
    Soilinputs.sndflg   =True
    Soilinputs.PenderSwtch   =False #True
    Soilinputs.SoilSF   =1.

    #Water and wind inputs
    Waterinputs=WaterInputs()
    Waterinputs.wdepth   =30.
    Waterinputs.wlevel   =30. #Distance from bottom of structure to surface  THIS, I believe is no longer needed as piles may be negative in z, to check and remove in case
    Waterinputs.T=12.  #Wave Period
    Waterinputs.HW=10. #Wave Height
    '''Windinputs=WindInputs()
    Windinputs.HH=100. #CHECK HOW THIS COMPLIES....
    Windinputs.U50HH=30. #assumed gust speed'''

    #RNA loads              Fx-z,         Mxx-zz
    #RNA_F=np.array([1000.e3,0.,0.,0.,0.,0.])

    #Pile data
    Pilematin=MatInputs()
    Pilematin.matname=np.array(['steel'])
    Pilematin.E=np.array([ 25.e9])
    Dpile=2.5#0.75 # 2.0
    tpile=0.01
    Lp=20. #45

    Pileinputs=PileGeoInputs()
    Pileinputs.Pilematins=Pilematin
    Pileinputs.ndiv=0 #3
    Pileinputs.Dpile=Dpile
    Pileinputs.tpile=tpile
    Pileinputs.Lp=Lp #[m] Embedment length

    #Legs data
    legmatin=MatInputs()
    legmatin.matname=(['steel','steel','steel','steel'])
    legmatin.E=np.array([2.0e11])
    Dleg=np.array([2.0,1.8,1.8,1.8,1.8,1.8])
    tleg=1.55*np.array([0.0254]).repeat(Dleg.size)
    leginputs=LegGeoInputs()
    leginputs.legZbot   = 1.0
    leginputs.ndiv=1
    leginputs.legmatins=legmatin
    leginputs.Dleg=Dleg
    leginputs.tleg=tleg

    legbot_stmphin =1.5  #Distance from bottom of leg to second joint along z; must be>0

    #Xbrc data
    Xbrcmatin=MatInputs()
    Xbrcmatin.matname=np.array(['steel']).repeat(Jcktins.nbays)
    Xbrcmatin.E=np.array([ 2.2e11, 2.0e11,2.0e11,2.0e11,2.0e11])
    Dbrc=np.array([1.,1.,1.0,1.0,1.0])
    tbrc=np.array([1.,1.,1.0,1.0,1.0])*0.0254

    Xbrcinputs=XBrcGeoInputs()
    Xbrcinputs.Dbrc=Dbrc
    Xbrcinputs.tbrc=tbrc
    Xbrcinputs.ndiv=2#2
    Xbrcinputs.Xbrcmatins=Xbrcmatin
    Xbrcinputs.precalc=True   #This can be set to true if we want Xbraces to be precalculated in D and t, in which case the above set Dbrc and tbrc would be overwritten

    #Mbrc data
    Mbrcmatin=MatInputs()
    Mbrcmatin.matname=np.array(['steel'])
    Mbrcmatin.E=np.array([ 2.5e11])
    Dbrc_mud=1.5

    Mbrcinputs=MudBrcGeoInputs()
    Mbrcinputs.Dbrc_mud=Dbrc_mud
    Mbrcinputs.ndiv=2
    Mbrcinputs.Mbrcmatins=Mbrcmatin
    Mbrcinputs.precalc=True   #This can be set to true if we want Mudbrace to be precalculated in D and t, in which case the above set Dbrc_mud and tbrc_mud would be overwritten
    #Hbrc data
    Hbrcmatin=MatInputs()
    Hbrcmatin.matname=np.array(['steel'])
    Hbrcmatin.E=np.array([ 2.5e11])
    Dbrc_hbrc=1.1

    Hbrcinputs=HBrcGeoInputs()
    Hbrcinputs.Dbrch=Dbrc_hbrc
    Hbrcinputs.ndiv=0#2
    Hbrcinputs.Hbrcmatins=Hbrcmatin
    Hbrcinputs.precalc=True   #This can be set to true if we want Hbrace to be set=Xbrace top D and t, in which case the above set Dbrch and tbrch would be overwritten

    #TP data
    TPlumpinputs=TPlumpMass()
    TPlumpinputs.mass=300.e3 #[kg]

    TPstmpsmatin=MatInputs()
    TPbrcmatin=MatInputs()
    TPstemmatin=MatInputs()
    TPbrcmatin.matname=np.array(['steel'])
    TPbrcmatin.E=np.array([ 2.5e11])
    TPstemmatin.matname=np.array(['steel']).repeat(2)
    TPstemmatin.E=np.array([ 2.1e11]).repeat(2)

    TPinputs=TPGeoInputs()
    TPinputs.TPbrcmatins=TPbrcmatin
    TPinputs.TPstemmatins=TPstemmatin
    TPinputs.TPstmpmatins=TPstmpsmatin
    TPinputs.Dstrut=1.6
    TPinputs.Dgir=Dbrc_hbrc
    TPinputs.Dbrc=1.1
    TPinputs.hstump=0.0#1.0
    TPinputs.stumpndiv=1#2
    TPinputs.brcndiv=1#2
    TPinputs.girndiv=1#2
    TPinputs.strutndiv=1#2
    TPinputs.stemndiv=1#2
    TPinputs.nstems=3
    TPinputs.Dstem=np.array([6.]).repeat(TPinputs.nstems)
    TPinputs.tstem=np.array([0.1,0.11,0.11])
    TPinputs.hstem=np.array([4.,3.,1.])

    #Tower data
    Twrmatin=MatInputs()
    Twrmatin.matname=np.array(['steel'])
    Twrmatin.E=np.array([ 2.77e11])
    Db=5.6
    tb=0.05
    Dt=Db*0.55

    '''Twrinputs=TwrGeoInputs()
    Twrinputs.Twrmatins=Twrmatin
    #Twrinputs.Htwr=70.  #Trumped by HH
    Twrinputs.Htwr2frac=0.2   #fraction of tower height with constant x-section
    Twrinputs.ndiv=np.array([6,6])  #ndiv for uniform and tapered section
    Twrinputs.Db=Db
    Twrinputs.DTRb=Db/tb
    #Twrinputs.Dt=Dt'''

    TwrRigidTop=True #False       #False=Account for RNA via math rather than a physical rigidmember

    #RNA data
    '''RNAins=RNAprops()
    RNAins.mass=3*350.e3
    RNAins.I[0]=86.579E+6
    RNAins.I[1]=53.530E+6
    RNAins.I[2]=58.112E+6
    RNAins.CMoff[2]=2.34
    RNAins.yawangle=45.  #angle with respect to global X, CCW looking from above, wind from left
    RNAins.rna_weightM=True'''

    #Frame3DD parameters
    FrameAuxIns=Frame3DDaux()
    FrameAuxIns.sh_fg=1               #shear flag-->Timoshenko
    FrameAuxIns.deltaz=5.
    FrameAuxIns.geo_fg=0
    FrameAuxIns.nModes = 6             # number of desired dynamic modes of vibration
    FrameAuxIns.Mmethod = 1            # 1: subspace Jacobi     2: Stodola
    FrameAuxIns.lump = 0               # 0: consistent mass ... 1: lumped mass matrix
    FrameAuxIns.tol = 1e-9             # mode shape tolerance
    FrameAuxIns.shift = 0.0            # shift value ... for unrestrained structures
    FrameAuxIns.gvector=np.array([0.,0.,-9.8065])    #GRAVITY

    #Pass all inputs to jacket assembly
    turbine.jacket.JcktGeoIn=Jcktins
    turbine.jacket.Soilinputs=Soilinputs
    turbine.jacket.Waterinputs=Waterinputs
    #turbine.jacket.Windinputs=Windinputs
    #turbine.jacket.RNA_F=RNA_F
    turbine.jacket.Pileinputs=Pileinputs
    turbine.jacket.leginputs=leginputs
    turbine.jacket.legbot_stmphin =legbot_stmphin
    turbine.jacket.Xbrcinputs=Xbrcinputs
    turbine.jacket.Mbrcinputs=Mbrcinputs
    turbine.jacket.Hbrcinputs=Hbrcinputs
    turbine.jacket.TPlumpinputs=TPlumpinputs
    turbine.jacket.TPinputs=TPinputs
    #turbine.jacket.RNAinputs=RNAins
    #turbine.jacket.Twrinputs=Twrinputs
    turbine.jacket.Twrinputs.Twrmatins=Twrmatin
    turbine.jacket.Twrinputs.Htwr2frac=0.2   #fraction of tower height with constant x-section
    turbine.jacket.Twrinputs.ndiv=np.array([6,6])  #ndiv for uniform and tapered section
    turbine.jacket.Twrinputs.Db=Db
    turbine.jacket.Twrinputs.Dt=Dt
    turbine.jacket.Twrinputs.DTRb=Db/tb
    turbine.jacket.Twrinputs.DTRt=Db/Dt # TODO double check
    turbine.jacket.TwrRigidTop=TwrRigidTop
    turbine.jacket.FrameAuxIns=FrameAuxIns




from __future__ import print_function
import numpy as np
from openmdao.api import ExplicitComponent, Group, IndepVarComp

from wisdem.commonse.utilities import hstack, vstack
from wisdem.commonse.csystem import DirectionVector
from wisdem.commonse import gravity

# This is an extremely simple RNA mass calculator that should be used when DriveSE otherwise seems too complicated


class RNAMass(ExplicitComponent):
    def setup(self):

        # variables
        self.add_input('blades_mass', 0.0, units='kg', desc='mass of all blade')
        self.add_input('hub_mass', 0.0, units='kg', desc='mass of hub')
        self.add_input('nac_mass', 0.0, units='kg', desc='mass of nacelle')

        self.add_input('hub_cm', np.zeros(3), units='m', desc='location of hub center of mass relative to tower top in yaw-aligned c.s.')
        self.add_input('nac_cm', np.zeros(3), units='m', desc='location of nacelle center of mass relative to tower top in yaw-aligned c.s.')

        # order for all moments of inertia is (xx, yy, zz, xy, xz, yz) in the yaw-aligned coorinate system
        self.add_input('blades_I', np.zeros(6), units='kg*m**2', desc='mass moments of inertia of all blades about hub center')
        self.add_input('hub_I', np.zeros(6), units='kg*m**2', desc='mass moments of inertia of hub about its center of mass')
        self.add_input('nac_I', np.zeros(6), units='kg*m**2', desc='mass moments of inertia of nacelle about its center of mass')

        # outputs
        self.add_output('rotor_mass', 0.0, units='kg', desc='mass of blades and hub')
        self.add_output('rna_mass', 0.0, units='kg', desc='total mass of RNA')
        self.add_output('rna_cm', np.zeros(3), units='m', desc='location of RNA center of mass relative to tower top in yaw-aligned c.s.')
        self.add_output('rna_I_TT', np.zeros(6), units='kg*m**2', desc='mass moments of inertia of RNA about tower top in yaw-aligned coordinate system')

        self.declare_partials('*','*')

    def _assembleI(self, I):
        Ixx, Iyy, Izz, Ixy, Ixz, Iyz = I[0], I[1], I[2], I[3], I[4], I[5] 
        return np.array([[Ixx, Ixy, Ixz], [Ixy, Iyy, Iyz], [Ixz, Iyz, Izz]])


    def _unassembleI(self, I):
        return np.array([I[0, 0], I[1, 1], I[2, 2], I[0, 1], I[0, 2], I[1, 2]])


    def compute(self, inputs, outputs):

        rotor_mass = inputs['blades_mass'] + inputs['hub_mass']
        nac_mass = inputs['nac_mass']

        # rna mass
        outputs['rotor_mass'] = rotor_mass
        outputs['rna_mass'] = rotor_mass + nac_mass

        # rna cm
        outputs['rna_cm'] = (rotor_mass*inputs['hub_cm'] + nac_mass*inputs['nac_cm'])/outputs['rna_mass']

        #TODO check if the use of assembleI and unassembleI functions are correct
        # rna I
        blades_I = self._assembleI(inputs['blades_I'])
        hub_I = self._assembleI(inputs['hub_I'])
        nac_I = self._assembleI(inputs['nac_I'])
        rotor_I = blades_I + hub_I

        R = inputs['hub_cm']
        rotor_I_TT = rotor_I + rotor_mass*(np.dot(R, R)*np.eye(3) - np.outer(R, R))

        R = inputs['nac_cm']
        nac_I_TT = nac_I + inputs['nac_mass']*(np.dot(R, R)*np.eye(3) - np.outer(R, R))

        outputs['rna_I_TT'] = self._unassembleI(rotor_I_TT + nac_I_TT)


    def compute_partials(self, inputs, J):

        blades_mass = inputs['blades_mass']
        hub_mass = inputs['hub_mass']
        nac_mass = inputs['nac_mass']
        hub_cm = inputs['hub_cm']
        nac_cm = inputs['nac_cm']
        hub_I = inputs['hub_I']
        nac_I = inputs['nac_I']
        rotor_mass = blades_mass+hub_mass
        rna_mass = rotor_mass + nac_mass

        

        # mass
        J['rotor_mass', 'blades_mass'] = 1.0
        J['rotor_mass', 'hub_mass'] = 1.0
        J['rotor_mass', 'nac_mass'] = 0.0
        J['rotor_mass', 'hub_cm'] = np.zeros(3)
        J['rotor_mass', 'nac_cm'] = np.zeros(3)
        J['rotor_mass', 'blades_I'] = np.zeros(6)
        J['rotor_mass', 'hub_I'] = np.zeros(6)
        J['rotor_mass', 'nac_I'] = np.zeros(6)

        J['rna_mass', 'blades_mass'] = 1.0
        J['rna_mass', 'hub_mass'] = 1.0
        J['rna_mass', 'nac_mass'] = 1.0
        J['rna_mass', 'hub_cm'] = np.zeros(3)
        J['rna_mass', 'nac_cm'] = np.zeros(3)
        J['rna_mass', 'blades_I'] = np.zeros(6)
        J['rna_mass', 'hub_I'] = np.zeros(6)
        J['rna_mass', 'nac_I'] = np.zeros(6)
        

        # cm
        numerator = (blades_mass+hub_mass)*hub_cm+nac_mass*nac_cm

        J['rna_cm', 'blades_mass'] = (rna_mass*hub_cm-numerator)/rna_mass**2
        J['rna_cm', 'hub_mass'] = (rna_mass*hub_cm-numerator)/rna_mass**2
        J['rna_cm', 'nac_mass'] = (rna_mass*nac_cm-numerator)/rna_mass**2
        J['rna_cm', 'hub_cm'] = rotor_mass/rna_mass*np.eye(3)
        J['rna_cm', 'nac_cm'] = nac_mass/rna_mass*np.eye(3)
        J['rna_cm', 'blades_I'] = np.zeros((3, 6))
        J['rna_cm', 'hub_I'] = np.zeros((3, 6))
        J['rna_cm', 'nac_I'] = np.zeros((3, 6))


        # I
        R = hub_cm
        const = self._unassembleI(np.dot(R, R)*np.eye(3) - np.outer(R, R))

        J['rna_I_TT', 'blades_mass'] = const
        J['rna_I_TT', 'hub_mass'] = const
        dI_drx = rotor_mass*self._unassembleI(2*R[0]*np.eye(3) - np.array([[2*R[0], R[1], R[2]], [R[1], 0.0, 0.0], [R[2], 0.0, 0.0]]))
        dI_dry = rotor_mass*self._unassembleI(2*R[1]*np.eye(3) - np.array([[0.0, R[0], 0.0], [R[0], 2*R[1], R[2]], [0.0, R[2], 0.0]]))
        dI_drz = rotor_mass*self._unassembleI(2*R[2]*np.eye(3) - np.array([[0.0, 0.0, R[0]], [0.0, 0.0, R[1]], [R[0], R[1], 2*R[2]]]))
        J['rna_I_TT', 'hub_cm'] = np.vstack([dI_drx, dI_dry, dI_drz]).T

        R = nac_cm
        const = self._unassembleI(np.dot(R, R)*np.eye(3) - np.outer(R, R))
        J['rna_I_TT', 'nac_mass'] = const
        dI_drx = nac_mass*self._unassembleI(2*R[0]*np.eye(3) - np.array([[2*R[0], R[1], R[2]], [R[1], 0.0, 0.0], [R[2], 0.0, 0.0]]))
        dI_dry = nac_mass*self._unassembleI(2*R[1]*np.eye(3) - np.array([[0.0, R[0], 0.0], [R[0], 2*R[1], R[2]], [0.0, R[2], 0.0]]))
        dI_drz = nac_mass*self._unassembleI(2*R[2]*np.eye(3) - np.array([[0.0, 0.0, R[0]], [0.0, 0.0, R[1]], [R[0], R[1], 2*R[2]]]))
        J['rna_I_TT', 'nac_cm'] = np.vstack([dI_drx, dI_dry, dI_drz]).T

        J['rna_I_TT', 'blades_I'] = np.eye(6)
        J['rna_I_TT', 'hub_I'] = np.eye(6)
        J['rna_I_TT', 'nac_I'] = np.eye(6)

        


class RotorLoads(ExplicitComponent):
    def setup(self):

        # variables
        self.add_input('F', np.zeros(3), units='N', desc='forces in hub-aligned coordinate system')
        self.add_input('M', np.zeros(3), units='N*m', desc='moments in hub-aligned coordinate system')
        self.add_input('hub_cm', np.zeros(3), units='m', desc='position of rotor hub relative to tower top in yaw-aligned c.s.')
        self.add_input('rna_mass', 0.0, units='kg', desc='mass of rotor nacelle assembly')
        self.add_input('rna_cm', np.zeros(3), units='m', desc='location of RNA center of mass relative to tower top in yaw-aligned c.s.')

        # # These are used for backwards compatibility - do not use
        # T = Float(iotype='in', desc='thrust in hub-aligned coordinate system')  # THIS MEANS STILL YAWED THOUGH (Shaft tilt)
        # Q = Float(iotype='in', desc='torque in hub-aligned coordinate system')

        # parameters
        self.add_discrete_input('downwind', False)
        self.add_input('tilt', 0.0, units='deg')

        # out
        self.add_output('top_F', np.zeros(3), units='N')  # in yaw-aligned
        self.add_output('top_M', np.zeros(3), units='N*m')

        self.declare_partials('top_F', ['F','M','hub_cm','rna_mass','rna_cm'])
        self.declare_partials('top_M', ['F','M','hub_cm','rna_mass','rna_cm'])


    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):

        F = inputs['F']
        M = inputs['M']
        tilt = float(inputs['tilt'])
        
        F = DirectionVector.fromArray(F).hubToYaw(tilt)
        M = DirectionVector.fromArray(M).hubToYaw(tilt)

        # change x-direction if downwind
        hub_cm = np.copy(inputs['hub_cm'])
        rna_cm = np.copy(inputs['rna_cm'])
        if discrete_inputs['downwind']:
            hub_cm[0] *= -1
            rna_cm[0] *= -1
        hub_cm = DirectionVector.fromArray(hub_cm)
        rna_cm = DirectionVector.fromArray(rna_cm)
        self.save_rhub = hub_cm
        self.save_rcm = rna_cm

        # aerodynamic moments
        M = M + hub_cm.cross(F)
        self.saveF = F

        '''
        Removing this permanently gbarter 1/2020 because of too much confusion in TowerSE and Frame3DD
        From now on TowerSE will always add to loading of added mass items, including RNA
        
        # add weight loads
        F_w = DirectionVector(0.0, 0.0, -float(inputs['rna_mass'])*gravity)
        M_w = rna_cm.cross(F_w)
        self.saveF_w = F_w

        Fout = F + F_w

        if discrete_inputs['rna_weightM']:
            Mout = M + M_w
        else:
            Mout = M
            #REMOVE WEIGHT EFFECT TO ACCOUNT FOR P-Delta Effect
            print("!!!! No weight effect on rotor moments -TowerSE  !!!!")
        '''
        Fout = F
        Mout = M

        # put back in array
        outputs['top_F'] = np.array([Fout.x, Fout.y, Fout.z])
        outputs['top_M'] = np.array([Mout.x, Mout.y, Mout.z])

    def compute_partials(self, inputs, J, discrete_inputs):

        dF = DirectionVector.fromArray(inputs['F']).hubToYaw(inputs['tilt'])
        dFx, dFy, dFz = dF.dx, dF.dy, dF.dz

        dtopF_dFx = np.array([dFx['dx'], dFy['dx'], dFz['dx']])
        dtopF_dFy = np.array([dFx['dy'], dFy['dy'], dFz['dy']])
        dtopF_dFz = np.array([dFx['dz'], dFy['dz'], dFz['dz']])
        dtopF_dF = hstack([dtopF_dFx, dtopF_dFy, dtopF_dFz])
        dtopF_w_dm = np.array([0.0, 0.0, -gravity])

        #dtopF = hstack([dtopF_dF, np.zeros((3, 6)), dtopF_w_dm, np.zeros((3, 3))])


        dM = DirectionVector.fromArray(inputs['M']).hubToYaw(inputs['tilt'])
        dMx, dMy, dMz = dM.dx, dM.dy, dM.dz
        dMxcross, dMycross, dMzcross = self.save_rhub.cross_deriv(self.saveF, 'dr', 'dF')

        dtopM_dMx = np.array([dMx['dx'], dMy['dx'], dMz['dx']])
        dtopM_dMy = np.array([dMx['dy'], dMy['dy'], dMz['dy']])
        dtopM_dMz = np.array([dMx['dz'], dMy['dz'], dMz['dz']])
        dtopM_dM = hstack([dtopM_dMx, dtopM_dMy, dtopM_dMz])
        dM_dF = np.array([dMxcross['dF'], dMycross['dF'], dMzcross['dF']])

        dtopM_dFx = np.dot(dM_dF, dtopF_dFx)
        dtopM_dFy = np.dot(dM_dF, dtopF_dFy)
        dtopM_dFz = np.dot(dM_dF, dtopF_dFz)
        dtopM_dF = hstack([dtopM_dFx, dtopM_dFy, dtopM_dFz])
        dtopM_dr = np.array([dMxcross['dr'], dMycross['dr'], dMzcross['dr']])

        #dMx_w_cross, dMy_w_cross, dMz_w_cross = self.save_rcm.cross_deriv(self.saveF_w, 'dr', 'dF')

        #if discrete_inputs['rna_weightM']:
        #    dtopM_drnacm = np.array([dMx_w_cross['dr'], dMy_w_cross['dr'], dMz_w_cross['dr']])
        #    dtopM_dF_w = np.array([dMx_w_cross['dF'], dMy_w_cross['dF'], dMz_w_cross['dF']])
        #else:
        #    dtopM_drnacm = np.zeros((3, 3))
        #    dtopM_dF_w = np.zeros((3, 3))
        dtopM_drnacm = np.zeros((3, 3))
        dtopM_dF_w = np.zeros((3, 3))
        dtopM_dm = np.dot(dtopM_dF_w, dtopF_w_dm)

        if discrete_inputs['downwind']:
            dtopM_dr[:, 0] *= -1
            dtopM_drnacm[:, 0] *= -1

        #dtopM = hstack([dtopM_dF, dtopM_dM, dtopM_dr, dtopM_dm, dtopM_drnacm])

        
        J['top_F', 'F'] = dtopF_dF
        J['top_F', 'M'] = np.zeros((3, 3))
        J['top_F', 'hub_cm'] = np.zeros((3, 3))
        J['top_F', 'rna_mass'] = dtopF_w_dm
        J['top_F', 'rna_cm'] = np.zeros((3, 3))

        J['top_M', 'F'] = dtopM_dF
        J['top_M', 'M'] = dtopM_dM
        J['top_M', 'hub_cm'] = dtopM_dr
        J['top_M', 'rna_mass'] = dtopM_dm
        J['top_M', 'rna_cm'] = dtopM_drnacm

        


class RNA(Group):
    def initialize(self):
        self.options.declare('nLC')
        
    def setup(self):
        nLC = self.options['nLC']
        
        self.add_subsystem('mass', RNAMass(), promotes=['*'])
        for k in range(nLC):
            lc = '' if nLC==1 else str(k+1)
            self.add_subsystem('loads'+lc, RotorLoads(), promotes=['rna_mass','rna_cm','hub_cm','downwind','tilt'])

        
